# -*- coding: utf-8 -*-
"""
Set up the version of Salt
"""
from __future__ import absolute_import, print_function, unicode_literals

import platform
import re
import sys

MAX_SIZE = sys.maxsize
VERSION_LIMIT = MAX_SIZE - 200

# ----- ATTENTION --------------------------------------------------------------------------------------------------->
#
# ALL major version bumps, new release codenames, MUST be defined in the SaltStackVersion.NAMES dictionary, i.e.:
#
#    class SaltStackVersion(object):
#
#        NAMES = {
#            'Hydrogen': (2014, 1),   # <- This is the tuple to bump versions
#            ( ... )
#        }
#
#
# ONLY UPDATE CODENAMES AFTER BRANCHING
#
# As an example, The Helium codename must only be properly defined with "(2014, 7)" after Hydrogen, "(2014, 1)", has
# been branched out into its own branch.
#
# ALL OTHER VERSION INFORMATION IS EXTRACTED FROM THE GIT TAGS
#
# <---- ATTENTION ----------------------------------------------------------------------------------------------------


class SaltStackVersion(object):
    """
    Handle SaltStack versions class.

    Knows how to parse ``git describe`` output, knows about release candidates
    and also supports version comparison.
    """

    __slots__ = (
        "name",
        "major",
        "minor",
        "bugfix",
        "mbugfix",
        "pre_type",
        "pre_num",
        "noc",
        "sha",
    )

    git_sha_regex = r"(?P<sha>g?[a-f0-9]{7,40})"

    git_describe_regex = re.compile(
        r"(?:[^\d]+)?(?P<major>[\d]{1,4})"
        r"(?:\.(?P<minor>[\d]{1,2}))?"
        r"(?:\.(?P<bugfix>[\d]{0,2}))?"
        r"(?:\.(?P<mbugfix>[\d]{0,2}))?"
        r"(?:(?P<pre_type>rc|a|b|alpha|beta|nb)(?P<pre_num>[\d]{1}))?"
        r"(?:(?:.*)-(?P<noc>(?:[\d]+|n/a))-" + git_sha_regex + r")?"
    )
    git_sha_regex = r"^" + git_sha_regex

    git_sha_regex = re.compile(git_sha_regex)

    # Salt versions after 0.17.0 will be numbered like:
    #   <4-digit-year>.<month>.<bugfix>
    #
    # Since the actual version numbers will only be know on release dates, the
    # periodic table element names will be what's going to be used to name
    # versions and to be able to mention them.

    NAMES = {
        # Let's keep at least 3 version names uncommented counting from the
        # latest release so we can map deprecation warnings to versions.
        # pylint: disable=E8203
        # ----- Please refrain from fixing PEP-8 E203 and E265 ----->
        # The idea is to keep this readable.
        # -----------------------------------------------------------
        "Hydrogen": (2014, 1),
        "Helium": (2014, 7),
        "Lithium": (2015, 5),
        "Beryllium": (2015, 8),
        "Boron": (2016, 3),
        "Carbon": (2016, 11),
        "Nitrogen": (2017, 7),
        "Oxygen": (2018, 3),
        "Fluorine": (2019, 2),
        "Neon": (3000,),
        "Sodium": (3001,),
        "Magnesium": (MAX_SIZE - 97, 0),
        "Aluminium": (MAX_SIZE - 96, 0),
        "Silicon": (MAX_SIZE - 95, 0),
        "Phosphorus": (MAX_SIZE - 94, 0),
        # pylint: disable=E8265
        #'Sulfur'       : (MAX_SIZE - 93, 0),
        #'Chlorine'     : (MAX_SIZE - 92, 0),
        #'Argon'        : (MAX_SIZE - 91, 0),
        #'Potassium'    : (MAX_SIZE - 90, 0),
        #'Calcium'      : (MAX_SIZE - 89, 0),
        #'Scandium'     : (MAX_SIZE - 88, 0),
        #'Titanium'     : (MAX_SIZE - 87, 0),
        #'Vanadium'     : (MAX_SIZE - 86, 0),
        #'Chromium'     : (MAX_SIZE - 85, 0),
        #'Manganese'    : (MAX_SIZE - 84, 0),
        #'Iron'         : (MAX_SIZE - 83, 0),
        #'Cobalt'       : (MAX_SIZE - 82, 0),
        #'Nickel'       : (MAX_SIZE - 81, 0),
        #'Copper'       : (MAX_SIZE - 80, 0),
        #'Zinc'         : (MAX_SIZE - 79, 0),
        #'Gallium'      : (MAX_SIZE - 78, 0),
        #'Germanium'    : (MAX_SIZE - 77, 0),
        #'Arsenic'      : (MAX_SIZE - 76, 0),
        #'Selenium'     : (MAX_SIZE - 75, 0),
        #'Bromine'      : (MAX_SIZE - 74, 0),
        #'Krypton'      : (MAX_SIZE - 73, 0),
        #'Rubidium'     : (MAX_SIZE - 72, 0),
        #'Strontium'    : (MAX_SIZE - 71, 0),
        #'Yttrium'      : (MAX_SIZE - 70, 0),
        #'Zirconium'    : (MAX_SIZE - 69, 0),
        #'Niobium'      : (MAX_SIZE - 68, 0),
        #'Molybdenum'   : (MAX_SIZE - 67, 0),
        #'Technetium'   : (MAX_SIZE - 66, 0),
        #'Ruthenium'    : (MAX_SIZE - 65, 0),
        #'Rhodium'      : (MAX_SIZE - 64, 0),
        #'Palladium'    : (MAX_SIZE - 63, 0),
        #'Silver'       : (MAX_SIZE - 62, 0),
        #'Cadmium'      : (MAX_SIZE - 61, 0),
        #'Indium'       : (MAX_SIZE - 60, 0),
        #'Tin'          : (MAX_SIZE - 59, 0),
        #'Antimony'     : (MAX_SIZE - 58, 0),
        #'Tellurium'    : (MAX_SIZE - 57, 0),
        #'Iodine'       : (MAX_SIZE - 56, 0),
        #'Xenon'        : (MAX_SIZE - 55, 0),
        #'Caesium'      : (MAX_SIZE - 54, 0),
        #'Barium'       : (MAX_SIZE - 53, 0),
        #'Lanthanum'    : (MAX_SIZE - 52, 0),
        #'Cerium'       : (MAX_SIZE - 51, 0),
        #'Praseodymium' : (MAX_SIZE - 50, 0),
        #'Neodymium'    : (MAX_SIZE - 49, 0),
        #'Promethium'   : (MAX_SIZE - 48, 0),
        #'Samarium'     : (MAX_SIZE - 47, 0),
        #'Europium'     : (MAX_SIZE - 46, 0),
        #'Gadolinium'   : (MAX_SIZE - 45, 0),
        #'Terbium'      : (MAX_SIZE - 44, 0),
        #'Dysprosium'   : (MAX_SIZE - 43, 0),
        #'Holmium'      : (MAX_SIZE - 42, 0),
        #'Erbium'       : (MAX_SIZE - 41, 0),
        #'Thulium'      : (MAX_SIZE - 40, 0),
        #'Ytterbium'    : (MAX_SIZE - 39, 0),
        #'Lutetium'     : (MAX_SIZE - 38, 0),
        #'Hafnium'      : (MAX_SIZE - 37, 0),
        #'Tantalum'     : (MAX_SIZE - 36, 0),
        #'Tungsten'     : (MAX_SIZE - 35, 0),
        #'Rhenium'      : (MAX_SIZE - 34, 0),
        #'Osmium'       : (MAX_SIZE - 33, 0),
        #'Iridium'      : (MAX_SIZE - 32, 0),
        #'Platinum'     : (MAX_SIZE - 31, 0),
        #'Gold'         : (MAX_SIZE - 30, 0),
        #'Mercury'      : (MAX_SIZE - 29, 0),
        #'Thallium'     : (MAX_SIZE - 28, 0),
        #'Lead'         : (MAX_SIZE - 27, 0),
        #'Bismuth'      : (MAX_SIZE - 26, 0),
        #'Polonium'     : (MAX_SIZE - 25, 0),
        #'Astatine'     : (MAX_SIZE - 24, 0),
        #'Radon'        : (MAX_SIZE - 23, 0),
        #'Francium'     : (MAX_SIZE - 22, 0),
        #'Radium'       : (MAX_SIZE - 21, 0),
        #'Actinium'     : (MAX_SIZE - 20, 0),
        #'Thorium'      : (MAX_SIZE - 19, 0),
        #'Protactinium' : (MAX_SIZE - 18, 0),
        #'Uranium'      : (MAX_SIZE - 17, 0),
        #'Neptunium'    : (MAX_SIZE - 16, 0),
        #'Plutonium'    : (MAX_SIZE - 15, 0),
        #'Americium'    : (MAX_SIZE - 14, 0),
        #'Curium'       : (MAX_SIZE - 13, 0),
        #'Berkelium'    : (MAX_SIZE - 12, 0),
        #'Californium'  : (MAX_SIZE - 11, 0),
        #'Einsteinium'  : (MAX_SIZE - 10, 0),
        #'Fermium'      : (MAX_SIZE - 9, 0),
        #'Mendelevium'  : (MAX_SIZE - 8, 0),
        #'Nobelium'     : (MAX_SIZE - 7, 0),
        #'Lawrencium'   : (MAX_SIZE - 6, 0),
        #'Rutherfordium': (MAX_SIZE - 5, 0),
        #'Dubnium'      : (MAX_SIZE - 4, 0),
        #'Seaborgium'   : (MAX_SIZE - 3, 0),
        #'Bohrium'      : (MAX_SIZE - 2, 0),
        #'Hassium'      : (MAX_SIZE - 1, 0),
        #'Meitnerium'   : (MAX_SIZE - 0, 0),
        # <---- Please refrain from fixing PEP-8 E203 and E265 ------
        # pylint: enable=E8203,E8265
    }

    LNAMES = dict((k.lower(), v) for (k, v) in iter(NAMES.items()))
    VNAMES = dict((v, k) for (k, v) in iter(NAMES.items()))
    RMATCH = dict((v[:2], k) for (k, v) in iter(NAMES.items()))

    def __init__(
        self,  # pylint: disable=C0103
        major,
        minor=None,
        bugfix=None,
        mbugfix=0,
        pre_type=None,
        pre_num=None,
        noc=0,
        sha=None,
    ):

        if isinstance(major, str):
            major = int(major)

        if isinstance(minor, str):
            if not minor:
                # Empty string
                minor = None
            else:
                minor = int(minor)

        if bugfix is None and not self.new_version(major=major):
            bugfix = 0
        elif isinstance(bugfix, str):
            if not bugfix:
                bugfix = None
            else:
                bugfix = int(bugfix)

        if mbugfix is None:
            mbugfix = 0
        elif isinstance(mbugfix, str):
            mbugfix = int(mbugfix)

        if pre_type is None:
            pre_type = ""
        if pre_num is None:
            pre_num = 0
        elif isinstance(pre_num, str):
            pre_num = int(pre_num)

        if noc is None:
            noc = 0
        elif isinstance(noc, str) and noc == "n/a":
            noc = -1
        elif isinstance(noc, str):
            noc = int(noc)

        self.major = major
        self.minor = minor
        self.bugfix = bugfix
        self.mbugfix = mbugfix
        self.pre_type = pre_type
        self.pre_num = pre_num
        self.name = self.VNAMES.get((major, minor), None)
        if self.new_version(major):
            self.name = self.VNAMES.get((major,), None)
        self.noc = noc
        self.sha = sha

    def new_version(self, major):
        """
        determine if using new versioning scheme
        """
        return bool(int(major) >= 3000 and int(major) < VERSION_LIMIT)

    @classmethod
    def parse(cls, version_string):
        if version_string.lower() in cls.LNAMES:
            return cls.from_name(version_string)
        vstr = (
            version_string.decode()
            if isinstance(version_string, bytes)
            else version_string
        )
        match = cls.git_describe_regex.match(vstr)
        if not match:
            raise ValueError(
                "Unable to parse version string: '{0}'".format(version_string)
            )
        return cls(*match.groups())

    @classmethod
    def from_name(cls, name):
        if name.lower() not in cls.LNAMES:
            raise ValueError("Named version '{0}' is not known".format(name))
        return cls(*cls.LNAMES[name.lower()])

    @classmethod
    def from_last_named_version(cls):
        return cls.from_name(
            cls.VNAMES[
                max(
                    [
                        version_info
                        for version_info in cls.VNAMES
                        if version_info[0] < (VERSION_LIMIT)
                    ]
                )
            ]
        )

    @classmethod
    def next_release(cls):
        return cls.from_name(
            cls.VNAMES[
                min(
                    [
                        version_info
                        for version_info in cls.VNAMES
                        if version_info > cls.from_last_named_version().info
                    ]
                )
            ]
        )

    @property
    def sse(self):
        # Higher than 0.17, lower than first date based
        return 0 < self.major < 2014

    def min_info(self):
        info = [self.major]
        if self.new_version(self.major):
            if self.minor:
                info.append(self.minor)
        else:
            info.extend([self.minor, self.bugfix, self.mbugfix])
        return info

    @property
    def info(self):
        return tuple(self.min_info())

    @property
    def pre_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num])
        return tuple(info)

    @property
    def noc_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num, self.noc])
        return tuple(info)

    @property
    def full_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num, self.noc, self.sha])
        return tuple(info)

    @property
    def full_info_all_versions(self):
        """
        Return the full info regardless
        of which versioning scheme we
        are using.
        """
        info = [
            self.major,
            self.minor,
            self.bugfix,
            self.mbugfix,
            self.pre_type,
            self.pre_num,
            self.noc,
            self.sha,
        ]
        return tuple(info)

    @property
    def string(self):
        if self.new_version(self.major):
            version_string = "{0}".format(self.major)
            if self.minor:
                version_string = "{0}.{1}".format(self.major, self.minor)
        else:
            version_string = "{0}.{1}.{2}".format(self.major, self.minor, self.bugfix)
        if self.mbugfix:
            version_string += ".{0}".format(self.mbugfix)
        if self.pre_type:
            version_string += "{0}{1}".format(self.pre_type, self.pre_num)
        if self.noc and self.sha:
            noc = self.noc
            if noc < 0:
                noc = "n/a"
            version_string += "-{0}-{1}".format(noc, self.sha)
        return version_string

    @property
    def formatted_version(self):
        if self.name and self.major > 10000:
            version_string = self.name
            if self.sse:
                version_string += " Enterprise"
            version_string += " (Unreleased)"
            return version_string
        version_string = self.string
        if self.sse:
            version_string += " Enterprise"
        if (self.major, self.minor) in self.RMATCH:
            version_string += " ({0})".format(self.RMATCH[(self.major, self.minor)])
        return version_string

    @property
    def pre_index(self):
        if self.new_version(self.major):
            pre_type = 2
            if not isinstance(self.minor, int):
                pre_type = 1
        else:
            pre_type = 4
        return pre_type

    def __str__(self):
        return self.string

    def __compare__(self, other, method):
        if not isinstance(other, SaltStackVersion):
            if isinstance(other, str):
                other = SaltStackVersion.parse(other)
            elif isinstance(other, (list, tuple)):
                other = SaltStackVersion(*other)
            else:
                raise ValueError(
                    "Cannot instantiate Version from type '{0}'".format(type(other))
                )

        pre_type = self.pre_index
        other_pre_type = other.pre_index
        other_noc_info = list(other.noc_info)
        noc_info = list(self.noc_info)

        if self.new_version(self.major):
            if self.minor and not other.minor:
                # We have minor information, the other side does not
                if self.minor > 0:
                    other_noc_info[1] = 0

            if not self.minor and other.minor:
                # The other side has minor information, we don't
                if other.minor > 0:
                    noc_info[1] = 0

        if self.pre_type and not other.pre_type:
            # We have pre-release information, the other side doesn't
            other_noc_info[other_pre_type] = "zzzzz"

        if not self.pre_type and other.pre_type:
            # The other side has pre-release information, we don't
            noc_info[pre_type] = "zzzzz"

        return method(tuple(noc_info), tuple(other_noc_info))

    def __lt__(self, other):
        return self.__compare__(other, lambda _self, _other: _self < _other)

    def __le__(self, other):
        return self.__compare__(other, lambda _self, _other: _self <= _other)

    def __eq__(self, other):
        return self.__compare__(other, lambda _self, _other: _self == _other)

    def __ne__(self, other):
        return self.__compare__(other, lambda _self, _other: _self != _other)

    def __ge__(self, other):
        return self.__compare__(other, lambda _self, _other: _self >= _other)

    def __gt__(self, other):
        return self.__compare__(other, lambda _self, _other: _self > _other)

    def __repr__(self):
        parts = []
        if self.name:
            parts.append("name='{0}'".format(self.name))
        parts.extend(["major={0}".format(self.major), "minor={0}".format(self.minor)])

        if self.new_version(self.major):
            if not self.minor:
                parts.remove("".join([x for x in parts if re.search("^minor*", x)]))
        else:
            parts.extend(["bugfix={0}".format(self.bugfix)])

        if self.mbugfix:
            parts.append("minor-bugfix={0}".format(self.mbugfix))
        if self.pre_type:
            parts.append("{0}={1}".format(self.pre_type, self.pre_num))
        noc = self.noc
        if noc == -1:
            noc = "n/a"
        if noc and self.sha:
            parts.extend(["noc={0}".format(noc), "sha={0}".format(self.sha)])
        return "<{0} {1}>".format(self.__class__.__name__, " ".join(parts))


# ----- Hardcoded Salt Codename Version Information ----------------------------------------------------------------->
#
#   There's no need to do anything here. The last released codename will be picked up
# --------------------------------------------------------------------------------------------------------------------
__saltstack_version__ = SaltStackVersion.from_last_named_version()
# <---- Hardcoded Salt Version Information ---------------------------------------------------------------------------


# ----- Dynamic/Runtime Salt Version Information -------------------------------------------------------------------->
def __discover_version(saltstack_version):
    # This might be a 'python setup.py develop' installation type. Let's
    # discover the version information at runtime.
    import os
    import subprocess

    if "SETUP_DIRNAME" in globals():
        # This is from the exec() call in Salt's setup.py
        cwd = SETUP_DIRNAME  # pylint: disable=E0602
        if not os.path.exists(os.path.join(cwd, ".git")):
            # This is not a Salt git checkout!!! Don't even try to parse...
            return saltstack_version
    else:
        cwd = os.path.abspath(os.path.dirname(__file__))
        if not os.path.exists(os.path.join(os.path.dirname(cwd), ".git")):
            # This is not a Salt git checkout!!! Don't even try to parse...
            return saltstack_version

    try:
        kwargs = dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)

        if not sys.platform.startswith("win"):
            # Let's not import `salt.utils` for the above check
            kwargs["close_fds"] = True

        process = subprocess.Popen(
            [
                "git",
                "describe",
                "--tags",
                "--first-parent",
                "--match",
                "v[0-9]*",
                "--always",
            ],
            **kwargs
        )

        out, err = process.communicate()

        if process.returncode != 0:
            # The git version running this might not support --first-parent
            # Revert to old command
            process = subprocess.Popen(
                ["git", "describe", "--tags", "--match", "v[0-9]*", "--always"],
                **kwargs
            )
            out, err = process.communicate()
        out = out.decode().strip()
        err = err.decode().strip()

        if not out or err:
            return saltstack_version

        if SaltStackVersion.git_sha_regex.match(out):
            # We only define the parsed SHA and set NOC as ??? (unknown)
            saltstack_version.sha = out.strip()
            saltstack_version.noc = -1
            return saltstack_version

        return SaltStackVersion.parse(out)

    except OSError as os_err:
        if os_err.errno != 2:
            # If the errno is not 2(The system cannot find the file
            # specified), raise the exception so it can be catch by the
            # developers
            raise
    return saltstack_version


def __get_version(saltstack_version):
    """
    If we can get a version provided at installation time or from Git, use
    that instead, otherwise we carry on.
    """
    try:
        # Try to import the version information provided at install time
        from salt._version import __saltstack_version__  # pylint: disable=E0611,F0401

        return __saltstack_version__
    except ImportError:
        return __discover_version(saltstack_version)


# Get additional version information if available
__saltstack_version__ = __get_version(__saltstack_version__)
# This function has executed once, we're done with it. Delete it!
del __get_version
# <---- Dynamic/Runtime Salt Version Information ---------------------------------------------------------------------


# ----- Common version related attributes - NO NEED TO CHANGE ------------------------------------------------------->
__version_info__ = __saltstack_version__.info
__version__ = __saltstack_version__.string
# <---- Common version related attributes - NO NEED TO CHANGE --------------------------------------------------------


def salt_information():
    """
    Report version of salt.
    """
    yield "Salt", __version__


def dependency_information(include_salt_cloud=False):
    """
    Report versions of library dependencies.
    """
    libs = [
        ("Python", None, sys.version.rsplit("\n")[0].strip()),
        ("Jinja2", "jinja2", "__version__"),
        ("M2Crypto", "M2Crypto", "version"),
        ("msgpack-python", "msgpack", "version"),
        ("msgpack-pure", "msgpack_pure", "version"),
        ("pycrypto", "Crypto", "__version__"),
        ("pycryptodome", "Cryptodome", "version_info"),
        ("PyYAML", "yaml", "__version__"),
        ("PyZMQ", "zmq", "__version__"),
        ("ZMQ", "zmq", "zmq_version"),
        ("Mako", "mako", "__version__"),
        ("Tornado", "tornado", "version"),
        ("timelib", "timelib", "version"),
        ("dateutil", "dateutil", "__version__"),
        ("pygit2", "pygit2", "__version__"),
        ("libgit2", "pygit2", "LIBGIT2_VERSION"),
        ("smmap", "smmap", "__version__"),
        ("cffi", "cffi", "__version__"),
        ("pycparser", "pycparser", "__version__"),
        ("gitdb", "gitdb", "__version__"),
        ("gitpython", "git", "__version__"),
        ("python-gnupg", "gnupg", "__version__"),
        ("mysql-python", "MySQLdb", "__version__"),
        ("cherrypy", "cherrypy", "__version__"),
        ("docker-py", "docker", "__version__"),
    ]

    if include_salt_cloud:
        libs.append(("Apache Libcloud", "libcloud", "__version__"),)

    for name, imp, attr in libs:
        if imp is None:
            yield name, attr
            continue
        try:
            imp = __import__(imp)
            version = getattr(imp, attr)
            if callable(version):
                version = version()
            if isinstance(version, (tuple, list)):
                version = ".".join(map(str, version))
            yield name, version
        except Exception:  # pylint: disable=broad-except
            yield name, None


def system_information():
    """
    Report system versions.
    """
    # Late import so that when getting called from setup.py does not break
    from distro import linux_distribution

    def system_version():
        """
        Return host system version.
        """

        lin_ver = linux_distribution()
        mac_ver = platform.mac_ver()
        win_ver = platform.win32_ver()

        if lin_ver[0]:
            return " ".join(lin_ver)
        elif mac_ver[0]:
            if isinstance(mac_ver[1], (tuple, list)) and "".join(mac_ver[1]):
                return " ".join([mac_ver[0], ".".join(mac_ver[1]), mac_ver[2]])
            else:
                return " ".join([mac_ver[0], mac_ver[2]])
        elif win_ver[0]:
            return " ".join(win_ver)
        else:
            return ""

    if platform.win32_ver()[0]:
        # Get the version and release info based on the Windows Operating
        # System Product Name. As long as Microsoft maintains a similar format
        # this should be future proof
        import win32api  # pylint: disable=3rd-party-module-not-gated
        import win32con  # pylint: disable=3rd-party-module-not-gated

        # Get the product name from the registry
        hkey = win32con.HKEY_LOCAL_MACHINE
        key = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
        value_name = "ProductName"
        reg_handle = win32api.RegOpenKey(hkey, key)

        # Returns a tuple of (product_name, value_type)
        product_name, _ = win32api.RegQueryValueEx(reg_handle, value_name)

        version = "Unknown"
        release = ""
        if "Server" in product_name:
            for item in product_name.split(" "):
                # If it's all digits, then it's version
                if re.match(r"\d+", item):
                    version = item
                # If it starts with R and then numbers, it's the release
                # ie: R2
                if re.match(r"^R\d+$", item):
                    release = item
            release = "{0}Server{1}".format(version, release)
        else:
            for item in product_name.split(" "):
                # If it's a number, decimal number, Thin or Vista, then it's the
                # version
                if re.match(r"^(\d+(\.\d+)?)|Thin|Vista$", item):
                    version = item
            release = version

        _, ver, service_pack, extra = platform.win32_ver()
        version = " ".join([release, ver, service_pack, extra])
    else:
        version = system_version()
        release = platform.release()

    system = [
        ("system", platform.system()),
        ("dist", " ".join(linux_distribution(full_distribution_name=False))),
        ("release", release),
        ("machine", platform.machine()),
        ("version", version),
        ("locale", __salt_system_encoding__),
    ]

    for name, attr in system:
        yield name, attr
        continue


def versions_information(include_salt_cloud=False):
    """
    Report the versions of dependent software.
    """
    salt_info = list(salt_information())
    lib_info = list(dependency_information(include_salt_cloud))
    sys_info = list(system_information())

    return {
        "Salt Version": dict(salt_info),
        "Dependency Versions": dict(lib_info),
        "System Versions": dict(sys_info),
    }


def versions_report(include_salt_cloud=False):
    """
    Yield each version properly formatted for console output.
    """
    ver_info = versions_information(include_salt_cloud)
    not_installed = "Not Installed"
    ns_pad = len(not_installed)
    lib_pad = max(len(name) for name in ver_info["Dependency Versions"])
    sys_pad = max(len(name) for name in ver_info["System Versions"])
    padding = max(lib_pad, sys_pad, ns_pad) + 1

    fmt = "{0:>{pad}}: {1}"
    info = []
    for ver_type in ("Salt Version", "Dependency Versions", "System Versions"):
        info.append("{0}:".format(ver_type))
        # List dependencies in alphabetical, case insensitive order
        for name in sorted(ver_info[ver_type], key=lambda x: x.lower()):
            ver = fmt.format(
                name, ver_info[ver_type][name] or not_installed, pad=padding
            )
            info.append(ver)
        info.append(" ")

    for line in info:
        yield line


if __name__ == "__main__":
    print(__version__)



# -*- coding: utf-8 -*-
"""
Tests for the archive state
"""
from __future__ import absolute_import, print_function, unicode_literals

import os
import shutil
import textwrap

import pytest
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt.ext import six
from tests.support.case import ModuleCase
from tests.support.helpers import destructiveTest, slowTest
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf

try:
    import zipfile  # pylint: disable=unused-import

    HAS_ZIPFILE = True
except ImportError:
    HAS_ZIPFILE = False


@destructiveTest
@pytest.mark.windows_whitelisted
class ArchiveTest(ModuleCase):
    """
    Validate the archive module
    """

    @classmethod
    def setUpClass(cls):
        # Base path used for test artifacts
        cls.base_path = os.path.join(RUNTIME_VARS.TMP, "modules", "archive")

    @classmethod
    def tearDownClass(cls):
        cls.base_path = None

    def _set_artifact_paths(self, arch_fmt):
        """
        Define the paths for the source, archive, and destination files

        :param str arch_fmt: The archive format used in the test
        """
        self.src = os.path.join(self.base_path, "{0}_src_dir".format(arch_fmt))
        self.src_file = os.path.join(self.src, "file")
        self.arch = os.path.join(self.base_path, "archive.{0}".format(arch_fmt))
        self.dst = os.path.join(self.base_path, "{0}_dst_dir".format(arch_fmt))

    def _set_up(self, arch_fmt, unicode_filename=False):
        """
        Create source file tree and destination directory

        :param str arch_fmt: The archive format used in the test
        """
        self._set_artifact_paths(arch_fmt)

        # Remove the artifacts if any present
        if any([os.path.exists(f) for f in (self.src, self.arch, self.dst)]):
            self._tear_down()
            self._set_artifact_paths(arch_fmt)

        # Create source
        os.makedirs(self.src)
        if unicode_filename:
            filename = "file®"
        else:
            filename = "file"
        with salt.utils.files.fopen(os.path.join(self.src, filename), "wb") as theorem:
            if six.PY3 and salt.utils.platform.is_windows():
                encoding = "utf-8"
            else:
                encoding = None
            theorem.write(
                salt.utils.stringutils.to_bytes(
                    textwrap.dedent(
                        """\
                Compression theorem of computational complexity theory:

                Given a Gödel numbering $φ$ of the computable functions and a
                Blum complexity measure $Φ$ where a complexity class for a
                boundary function $f$ is defined as

                    $\\mathrm C(f) := \\{φ_i ∈ \\mathbb R^{(1)} | (∀^∞ x) Φ_i(x) ≤ f(x)\\}$.

                Then there exists a total computable function $f$ so that for
                all $i$

                    $\\mathrm{Dom}(φ_i) = \\mathrm{Dom}(φ_{f(i)})$

                and

                    $\\mathrm C(φ_i) ⊊ \\mathrm{C}(φ_{f(i)})$.
            """
                    ),
                    encoding=encoding,
                )
            )

        # Create destination
        os.makedirs(self.dst)

    def _tear_down(self):
        """
        Remove source file tree, archive, and destination file tree
        """
        for f in (self.src, self.arch, self.dst):
            if os.path.exists(f):
                if os.path.isdir(f):
                    shutil.rmtree(f, ignore_errors=True)
                else:
                    os.remove(f)
        del self.dst
        del self.src
        del self.arch
        del self.src_file

    def _assert_artifacts_in_ret(self, ret, file_only=False, unix_sep=False):
        """
        Assert that the artifact source files are printed in the source command
        output
        """

        def normdir(path):
            normdir = os.path.normcase(os.path.abspath(path))
            if salt.utils.platform.is_windows():
                # Remove the drive portion of path
                if len(normdir) >= 2 and normdir[1] == ":":
                    normdir = normdir.split(":", 1)[1]
            normdir = normdir.lstrip(os.path.sep)
            # Unzipped paths might have unix line endings
            if unix_sep:
                normdir = normdir.replace(os.path.sep, "/")
            return normdir

        # Try to find source directory and file in output lines
        dir_in_ret = None
        file_in_ret = None
        for line in ret:
            if normdir(self.src) in os.path.normcase(line) and not normdir(
                self.src_file
            ) in os.path.normcase(line):
                dir_in_ret = True
            if normdir(self.src_file) in os.path.normcase(line):
                file_in_ret = True

        # Assert number of lines, reporting of source directory and file
        self.assertTrue(len(ret) >= 1 if file_only else 2)
        if not file_only:
            self.assertTrue(dir_in_ret)
        self.assertTrue(file_in_ret)

    @skipIf(not salt.utils.path.which("tar"), "Cannot find tar executable")
    @slowTest
    def test_tar_pack(self):
        """
        Validate using the tar function to create archives
        """
        self._set_up(arch_fmt="tar")

        # Test create archive
        ret = self.run_function("archive.tar", ["-cvf", self.arch], sources=self.src)
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not salt.utils.path.which("tar"), "Cannot find tar executable")
    @slowTest
    def test_tar_unpack(self):
        """
        Validate using the tar function to extract archives
        """
        self._set_up(arch_fmt="tar")
        self.run_function("archive.tar", ["-cvf", self.arch], sources=self.src)

        # Test extract archive
        ret = self.run_function("archive.tar", ["-xvf", self.arch], dest=self.dst)
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not salt.utils.path.which("tar"), "Cannot find tar executable")
    @slowTest
    def test_tar_pack_unicode(self):
        """
        Validate using the tar function to create archives
        """
        self._set_up(arch_fmt="tar", unicode_filename=True)

        # Test create archive
        ret = self.run_function("archive.tar", ["-cvf", self.arch], sources=self.src)
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not salt.utils.path.which("tar"), "Cannot find tar executable")
    @slowTest
    def test_tar_unpack_unicode(self):
        """
        Validate using the tar function to extract archives
        """
        self._set_up(arch_fmt="tar", unicode_filename=True)
        self.run_function("archive.tar", ["-cvf", self.arch], sources=self.src)

        # Test extract archive
        ret = self.run_function("archive.tar", ["-xvf", self.arch], dest=self.dst)
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not salt.utils.path.which("tar"), "Cannot find tar executable")
    @slowTest
    def test_tar_list_unicode(self):
        """
        Validate using the tar function to extract archives
        """
        self._set_up(arch_fmt="tar", unicode_filename=True)
        self.run_function("archive.tar", ["-cvf", self.arch], sources=self.src)

        # Test list archive
        ret = self.run_function("archive.list", name=self.arch)
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not salt.utils.path.which("gzip"), "Cannot find gzip executable")
    def test_gzip(self):
        """
        Validate using the gzip function
        """
        self._set_up(arch_fmt="gz")

        # Test create archive
        ret = self.run_function("archive.gzip", [self.src_file], options="-v")
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret, file_only=True)

        self._tear_down()

    @skipIf(not salt.utils.path.which("gzip"), "Cannot find gzip executable")
    @skipIf(not salt.utils.path.which("gunzip"), "Cannot find gunzip executable")
    def test_gunzip(self):
        """
        Validate using the gunzip function
        """
        self._set_up(arch_fmt="gz")
        self.run_function("archive.gzip", [self.src_file], options="-v")

        # Test extract archive
        ret = self.run_function("archive.gunzip", [self.src_file + ".gz"], options="-v")
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret, file_only=True)

        self._tear_down()

    @skipIf(not salt.utils.path.which("zip"), "Cannot find zip executable")
    def test_cmd_zip(self):
        """
        Validate using the cmd_zip function
        """
        self._set_up(arch_fmt="zip")

        # Test create archive
        ret = self.run_function("archive.cmd_zip", [self.arch, self.src])
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not salt.utils.path.which("zip"), "Cannot find zip executable")
    @skipIf(not salt.utils.path.which("unzip"), "Cannot find unzip executable")
    def test_cmd_unzip(self):
        """
        Validate using the cmd_unzip function
        """
        self._set_up(arch_fmt="zip")
        self.run_function("archive.cmd_zip", [self.arch, self.src])

        # Test create archive
        ret = self.run_function("archive.cmd_unzip", [self.arch, self.dst])
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not HAS_ZIPFILE, "Cannot find zipfile python module")
    @slowTest
    def test_zip(self):
        """
        Validate using the zip function
        """
        self._set_up(arch_fmt="zip")

        # Test create archive
        ret = self.run_function("archive.zip", [self.arch, self.src])
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not HAS_ZIPFILE, "Cannot find zipfile python module")
    @slowTest
    def test_unzip(self):
        """
        Validate using the unzip function
        """
        self._set_up(arch_fmt="zip")
        self.run_function("archive.zip", [self.arch, self.src])

        # Test create archive
        ret = self.run_function("archive.unzip", [self.arch, self.dst])
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret, unix_sep=False)

        self._tear_down()

    @skipIf(not salt.utils.path.which("rar"), "Cannot find rar executable")
    def test_rar(self):
        """
        Validate using the rar function
        """
        self._set_up(arch_fmt="rar")

        # Test create archive
        ret = self.run_function("archive.rar", [self.arch, self.src])
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()

    @skipIf(not salt.utils.path.which("rar"), "Cannot find rar executable")
    @skipIf(not salt.utils.path.which("unrar"), "Cannot find unrar executable")
    def test_unrar(self):
        """
        Validate using the unrar function
        """
        self._set_up(arch_fmt="rar")
        self.run_function("archive.rar", [self.arch, self.src])

        # Test create archive
        ret = self.run_function("archive.unrar", [self.arch, self.dst])
        self.assertTrue(isinstance(ret, list), six.text_type(ret))
        self._assert_artifacts_in_ret(ret)

        self._tear_down()



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

import os

# Import Salt Libs
import salt.modules.win_path as win_path
import salt.utils.stringutils

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class WinPathTestCase(TestCase, LoaderModuleMockMixin):
    """
        Test cases for salt.modules.win_path
    """

    def setup_loader_modules(self):
        return {win_path: {}}

    def __init__(self, *args, **kwargs):
        super(WinPathTestCase, self).__init__(*args, **kwargs)
        self.pathsep = str(";")  # future lint: disable=blacklisted-function

    def assert_call_matches(self, mock_obj, new_path):
        mock_obj.assert_called_once_with(
            win_path.HIVE,
            win_path.KEY,
            win_path.VNAME,
            self.pathsep.join(new_path),
            win_path.VTYPE,
        )

    def assert_path_matches(self, env, new_path):
        self.assertEqual(
            env["PATH"], salt.utils.stringutils.to_str(self.pathsep.join(new_path))
        )

    def test_get_path(self):
        """
        Test to Returns the system path
        """
        mock = MagicMock(return_value={"vdata": "C:\\Salt"})
        with patch.dict(win_path.__utils__, {"reg.read_value": mock}):
            self.assertListEqual(win_path.get_path(), ["C:\\Salt"])

    def test_exists(self):
        """
        Test to check if the directory is configured
        """
        get_mock = MagicMock(return_value=["C:\\Foo", "C:\\Bar"])
        with patch.object(win_path, "get_path", get_mock):
            # Ensure case insensitivity respected
            self.assertTrue(win_path.exists("C:\\FOO"))
            self.assertTrue(win_path.exists("c:\\foo"))
            self.assertFalse(win_path.exists("c:\\mystuff"))

    def test_add(self):
        """
        Test to add the directory to the SYSTEM path
        """
        orig_path = ("C:\\Foo", "C:\\Bar")

        def _env(path):
            return {
                str(
                    "PATH"
                ): salt.utils.stringutils.to_str(  # future lint: disable=blacklisted-function
                    self.pathsep.join(path)
                )
            }

        def _run(name, index=None, retval=True, path=None):
            if path is None:
                path = orig_path
            env = _env(path)
            mock_get = MagicMock(return_value=list(path))
            mock_set = MagicMock(return_value=retval)
            with patch.object(win_path, "PATHSEP", self.pathsep), patch.object(
                win_path, "get_path", mock_get
            ), patch.object(os, "environ", env), patch.dict(
                win_path.__utils__, {"reg.set_value": mock_set}
            ), patch.object(
                win_path, "rehash", MagicMock(return_value=True)
            ):
                return win_path.add(name, index), env, mock_set

        # Test a successful reg update
        ret, env, mock_set = _run("c:\\salt", retval=True)
        new_path = ("C:\\Foo", "C:\\Bar", "c:\\salt")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test an unsuccessful reg update
        ret, env, mock_set = _run("c:\\salt", retval=False)
        new_path = ("C:\\Foo", "C:\\Bar", "c:\\salt")
        self.assertFalse(ret)
        self.assert_call_matches(mock_set, new_path)
        # The local path should still have been modified even
        # though reg.set_value failed.
        self.assert_path_matches(env, new_path)

        # Test adding with a custom index
        ret, env, mock_set = _run("c:\\salt", index=1, retval=True)
        new_path = ("C:\\Foo", "c:\\salt", "C:\\Bar")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test adding with a custom index of 0
        ret, env, mock_set = _run("c:\\salt", index=0, retval=True)
        new_path = ("c:\\salt", "C:\\Foo", "C:\\Bar")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test adding path with a case-insensitive match already present, and
        # no index provided. The path should remain unchanged and we should not
        # update the registry.
        ret, env, mock_set = _run("c:\\foo", retval=True)
        self.assertTrue(ret)
        mock_set.assert_not_called()
        self.assert_path_matches(env, orig_path)

        # Test adding path with a case-insensitive match already present, and a
        # negative index provided which does not match the current index. The
        # match should be removed, and the path should be added to the end of
        # the list.
        ret, env, mock_set = _run("c:\\foo", index=-1, retval=True)
        new_path = ("C:\\Bar", "c:\\foo")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test adding path with a case-insensitive match already present, and a
        # negative index provided which matches the current index. No changes
        # should be made.
        ret, env, mock_set = _run("c:\\foo", index=-2, retval=True)
        self.assertTrue(ret)
        mock_set.assert_not_called()
        self.assert_path_matches(env, orig_path)

        # Test adding path with a case-insensitive match already present, and a
        # negative index provided which is larger than the size of the list. No
        # changes should be made, since in these cases we assume an index of 0,
        # and the case-insensitive match is also at index 0.
        ret, env, mock_set = _run("c:\\foo", index=-5, retval=True)
        self.assertTrue(ret)
        mock_set.assert_not_called()
        self.assert_path_matches(env, orig_path)

        # Test adding path with a case-insensitive match already present, and a
        # negative index provided which is larger than the size of the list.
        # The match should be removed from its current location and inserted at
        # the beginning, since when a negative index is larger than the list,
        # we put it at the beginning of the list.
        ret, env, mock_set = _run("c:\\bar", index=-5, retval=True)
        new_path = ("c:\\bar", "C:\\Foo")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test adding path with a case-insensitive match already present, and a
        # negative index provided which matches the current index. The path
        # should remain unchanged and we should not update the registry.
        ret, env, mock_set = _run("c:\\bar", index=-1, retval=True)
        self.assertTrue(ret)
        mock_set.assert_not_called()
        self.assert_path_matches(env, orig_path)

        # Test adding path with a case-insensitive match already present, and
        # an index provided which does not match the current index, and is also
        # larger than the size of the PATH list. The match should be removed,
        # and the path should be added to the end of the list.
        ret, env, mock_set = _run("c:\\foo", index=5, retval=True)
        new_path = ("C:\\Bar", "c:\\foo")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

    def test_remove(self):
        """
        Test win_path.remove
        """
        orig_path = ("C:\\Foo", "C:\\Bar", "C:\\Baz")

        def _env(path):
            return {
                str(
                    "PATH"
                ): salt.utils.stringutils.to_str(  # future lint: disable=blacklisted-function
                    self.pathsep.join(path)
                )
            }

        def _run(name="c:\\salt", index=None, retval=True, path=None):
            if path is None:
                path = orig_path
            env = _env(path)
            mock_get = MagicMock(return_value=list(path))
            mock_set = MagicMock(return_value=retval)
            with patch.object(win_path, "PATHSEP", self.pathsep), patch.object(
                win_path, "get_path", mock_get
            ), patch.object(os, "environ", env), patch.dict(
                win_path.__utils__, {"reg.set_value": mock_set}
            ), patch.object(
                win_path, "rehash", MagicMock(return_value=True)
            ):
                return win_path.remove(name), env, mock_set

        # Test a successful reg update
        ret, env, mock_set = _run("C:\\Bar", retval=True)
        new_path = ("C:\\Foo", "C:\\Baz")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test a successful reg update with a case-insensitive match
        ret, env, mock_set = _run("c:\\bar", retval=True)
        new_path = ("C:\\Foo", "C:\\Baz")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test a successful reg update with multiple case-insensitive matches.
        # All matches should be removed.
        old_path = orig_path + ("C:\\BAR",)
        ret, env, mock_set = _run("c:\\bar", retval=True)
        new_path = ("C:\\Foo", "C:\\Baz")
        self.assertTrue(ret)
        self.assert_call_matches(mock_set, new_path)
        self.assert_path_matches(env, new_path)

        # Test an unsuccessful reg update
        ret, env, mock_set = _run("c:\\bar", retval=False)
        new_path = ("C:\\Foo", "C:\\Baz")
        self.assertFalse(ret)
        self.assert_call_matches(mock_set, new_path)
        # The local path should still have been modified even
        # though reg.set_value failed.
        self.assert_path_matches(env, new_path)

        # Test when no match found
        ret, env, mock_set = _run("C:\\NotThere", retval=True)
        self.assertTrue(ret)
        mock_set.assert_not_called()
        self.assert_path_matches(env, orig_path)



# -*- coding: utf-8 -*-
"""
Create ssh executor system
"""
from __future__ import absolute_import, print_function

# Import python libs
import logging
import os
import shutil
import tarfile
import tempfile
from contextlib import closing

import salt.client.ssh

# Import salt libs
import salt.client.ssh.shell
import salt.loader
import salt.minion
import salt.roster
import salt.state
import salt.utils.files
import salt.utils.json
import salt.utils.path
import salt.utils.stringutils
import salt.utils.thin
import salt.utils.url
import salt.utils.verify

# Import 3rd-party libs
from salt.ext import six

log = logging.getLogger(__name__)


class SSHState(salt.state.State):
    """
    Create a State object which wraps the SSH functions for state operations
    """

    def __init__(self, opts, pillar=None, wrapper=None):
        self.wrapper = wrapper
        super(SSHState, self).__init__(opts, pillar)

    def load_modules(self, data=None, proxy=None):
        """
        Load up the modules for remote compilation via ssh
        """
        self.functions = self.wrapper
        self.utils = salt.loader.utils(self.opts)
        self.serializers = salt.loader.serializers(self.opts)
        locals_ = salt.loader.minion_mods(self.opts, utils=self.utils)
        self.states = salt.loader.states(
            self.opts, locals_, self.utils, self.serializers
        )
        self.rend = salt.loader.render(self.opts, self.functions)

    def check_refresh(self, data, ret):
        """
        Stub out check_refresh
        """
        return

    def module_refresh(self):
        """
        Module refresh is not needed, stub it out
        """
        return


class SSHHighState(salt.state.BaseHighState):
    """
    Used to compile the highstate on the master
    """

    stack = []

    def __init__(self, opts, pillar=None, wrapper=None, fsclient=None):
        self.client = fsclient
        salt.state.BaseHighState.__init__(self, opts)
        self.state = SSHState(opts, pillar, wrapper)
        self.matchers = salt.loader.matchers(self.opts)
        self.tops = salt.loader.tops(self.opts)

        self._pydsl_all_decls = {}
        self._pydsl_render_stack = []

    def push_active(self):
        salt.state.HighState.stack.append(self)

    def load_dynamic(self, matches):
        """
        Stub out load_dynamic
        """
        return

    def _master_tops(self):
        """
        Evaluate master_tops locally
        """
        if "id" not in self.opts:
            log.error("Received call for external nodes without an id")
            return {}
        if not salt.utils.verify.valid_id(self.opts, self.opts["id"]):
            return {}
        # Evaluate all configured master_tops interfaces

        grains = {}
        ret = {}

        if "grains" in self.opts:
            grains = self.opts["grains"]
        for fun in self.tops:
            if fun not in self.opts.get("master_tops", {}):
                continue
            try:
                ret.update(self.tops[fun](opts=self.opts, grains=grains))
            except Exception as exc:  # pylint: disable=broad-except
                # If anything happens in the top generation, log it and move on
                log.error(
                    "Top function %s failed with error %s for minion %s",
                    fun,
                    exc,
                    self.opts["id"],
                )
        return ret


def lowstate_file_refs(chunks, extras=""):
    """
    Create a list of file ref objects to reconcile
    """
    refs = {}
    for chunk in chunks:
        if not isinstance(chunk, dict):
            continue
        saltenv = "base"
        crefs = []
        for state in chunk:
            if state == "__env__":
                saltenv = chunk[state]
            elif state.startswith("__"):
                continue
            crefs.extend(salt_refs(chunk[state]))
        if saltenv not in refs:
            refs[saltenv] = []
        if crefs:
            refs[saltenv].append(crefs)
    if extras:
        extra_refs = extras.split(",")
        if extra_refs:
            for env in refs:
                for x in extra_refs:
                    refs[env].append([x])
    return refs


def salt_refs(data, ret=None):
    """
    Pull salt file references out of the states
    """
    proto = "salt://"
    if ret is None:
        ret = []
    if isinstance(data, six.string_types):
        if data.startswith(proto) and data not in ret:
            ret.append(data)
    if isinstance(data, list):
        for comp in data:
            salt_refs(comp, ret)
    if isinstance(data, dict):
        for comp in data:
            salt_refs(data[comp], ret)
    return ret


def prep_trans_tar(
    file_client, chunks, file_refs, pillar=None, id_=None, roster_grains=None
):
    """
    Generate the execution package from the saltenv file refs and a low state
    data structure
    """
    gendir = tempfile.mkdtemp()
    trans_tar = salt.utils.files.mkstemp()
    lowfn = os.path.join(gendir, "lowstate.json")
    pillarfn = os.path.join(gendir, "pillar.json")
    roster_grainsfn = os.path.join(gendir, "roster_grains.json")
    sync_refs = [
        [salt.utils.url.create("_modules")],
        [salt.utils.url.create("_states")],
        [salt.utils.url.create("_grains")],
        [salt.utils.url.create("_renderers")],
        [salt.utils.url.create("_returners")],
        [salt.utils.url.create("_output")],
        [salt.utils.url.create("_utils")],
    ]
    with salt.utils.files.fopen(lowfn, "w+") as fp_:
        salt.utils.json.dump(chunks, fp_)
    if pillar:
        with salt.utils.files.fopen(pillarfn, "w+") as fp_:
            salt.utils.json.dump(pillar, fp_)
    if roster_grains:
        with salt.utils.files.fopen(roster_grainsfn, "w+") as fp_:
            salt.utils.json.dump(roster_grains, fp_)

    if id_ is None:
        id_ = ""
    try:
        cachedir = os.path.join("salt-ssh", id_).rstrip(os.sep)
    except AttributeError:
        # Minion ID should always be a str, but don't let an int break this
        cachedir = os.path.join("salt-ssh", six.text_type(id_)).rstrip(os.sep)

    for saltenv in file_refs:
        # Location where files in this saltenv will be cached
        cache_dest_root = os.path.join(cachedir, "files", saltenv)
        file_refs[saltenv].extend(sync_refs)
        env_root = os.path.join(gendir, saltenv)
        if not os.path.isdir(env_root):
            os.makedirs(env_root)
        for ref in file_refs[saltenv]:
            for name in ref:
                short = salt.utils.url.parse(name)[0].lstrip("/")
                cache_dest = os.path.join(cache_dest_root, short)
                try:
                    path = file_client.cache_file(name, saltenv, cachedir=cachedir)
                except IOError:
                    path = ""
                if path:
                    tgt = os.path.join(env_root, short)
                    tgt_dir = os.path.dirname(tgt)
                    if not os.path.isdir(tgt_dir):
                        os.makedirs(tgt_dir)
                    shutil.copy(path, tgt)
                    continue
                try:
                    files = file_client.cache_dir(name, saltenv, cachedir=cachedir)
                except IOError:
                    files = ""
                if files:
                    for filename in files:
                        fn = filename[
                            len(file_client.get_cachedir(cache_dest)) :
                        ].strip("/")
                        tgt = os.path.join(env_root, short, fn,)
                        tgt_dir = os.path.dirname(tgt)
                        if not os.path.isdir(tgt_dir):
                            os.makedirs(tgt_dir)
                        shutil.copy(filename, tgt)
                    continue
    try:
        # cwd may not exist if it was removed but salt was run from it
        cwd = os.getcwd()
    except OSError:
        cwd = None
    os.chdir(gendir)
    with closing(tarfile.open(trans_tar, "w:gz")) as tfp:
        for root, dirs, files in salt.utils.path.os_walk(gendir):
            for name in files:
                full = os.path.join(root, name)
                tfp.add(full[len(gendir) :].lstrip(os.sep))
    if cwd:
        os.chdir(cwd)
    shutil.rmtree(gendir)
    return trans_tar



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import logging

# Import Salt Libs
import salt.states.host as host

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class HostTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the host state
    """

    hostname = "salt"
    localhost_ip = "127.0.0.1"
    ip_list = ["203.0.113.113", "203.0.113.14"]
    default_hosts = {
        ip_list[0]: {"aliases": [hostname]},
        ip_list[1]: {"aliases": [hostname]},
    }

    def setUp(self):
        self.add_host_mock = MagicMock(return_value=True)
        self.rm_host_mock = MagicMock(return_value=True)
        self.set_comment_mock = MagicMock(return_value=True)
        self.list_hosts_mock = MagicMock(return_value=self.default_hosts)

    def setup_loader_modules(self):
        return {
            host: {"__opts__": {"test": False}},
        }

    def test_present(self):
        """
        Test to ensures that the named host is present with the given ip
        """
        add_host = MagicMock(return_value=True)
        rm_host = MagicMock(return_value=True)
        set_comment = MagicMock(return_value=True)
        hostname = "salt"
        ip_str = "127.0.0.1"
        ip_list = ["10.1.2.3", "10.4.5.6"]

        # Case 1: No match for hostname. Single IP address passed to the state.
        list_hosts = MagicMock(return_value={"127.0.0.1": {"aliases": ["localhost"]}})
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_str)
            assert ret["result"] is True
            assert ret["comment"] == "Added host {0} ({1})".format(
                hostname, ip_str
            ), ret["comment"]
            assert ret["changes"] == {"added": {ip_str: [hostname]}}, ret["changes"]
            expected = [call(ip_str, hostname)]
            assert add_host.mock_calls == expected, add_host.mock_calls
            assert rm_host.mock_calls == [], rm_host.mock_calls

        # Case 2: No match for hostname. Multiple IP addresses passed to the
        # state.
        list_hosts = MagicMock(return_value={"127.0.0.1": {"aliases": ["localhost"]}})
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_list)
            assert ret["result"] is True
            assert "Added host {0} ({1})".format(hostname, ip_list[0]) in ret["comment"]
            assert "Added host {0} ({1})".format(hostname, ip_list[1]) in ret["comment"]
            assert ret["changes"] == {
                "added": {ip_list[0]: [hostname], ip_list[1]: [hostname]}
            }, ret["changes"]
            expected = sorted([call(x, hostname) for x in ip_list])
            assert sorted(add_host.mock_calls) == expected, add_host.mock_calls
            assert rm_host.mock_calls == [], rm_host.mock_calls

        # Case 3: Match for hostname, but no matching IP. Single IP address
        # passed to the state.
        list_hosts = MagicMock(
            return_value={
                "127.0.0.1": {"aliases": ["localhost"]},
                ip_list[0]: {"aliases": [hostname]},
            }
        )
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_str)
            assert ret["result"] is True
            assert "Added host {0} ({1})".format(hostname, ip_str) in ret["comment"]
            assert (
                "Host {0} present for IP address {1}".format(hostname, ip_list[0])
                in ret["warnings"][0]
            )
            assert ret["changes"] == {"added": {ip_str: [hostname]}}, ret["changes"]
            expected = [call(ip_str, hostname)]
            assert add_host.mock_calls == expected, add_host.mock_calls
            assert rm_host.mock_calls == [], rm_host.mock_calls

        # Case 3a: Repeat the above with clean=True
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_str, clean=True)
            assert ret["result"] is True
            assert "Added host {0} ({1})".format(hostname, ip_str) in ret["comment"]
            assert (
                "Removed host {0} ({1})".format(hostname, ip_list[0]) in ret["comment"]
            )
            assert ret["changes"] == {
                "added": {ip_str: [hostname]},
                "removed": {ip_list[0]: [hostname]},
            }, ret["changes"]
            expected = [call(ip_str, hostname)]
            assert add_host.mock_calls == expected, add_host.mock_calls
            expected = [call(ip_list[0], hostname)]
            assert rm_host.mock_calls == expected, rm_host.mock_calls

        # Case 4: Match for hostname, but no matching IP. Multiple IP addresses
        # passed to the state.
        cur_ip = "1.2.3.4"
        list_hosts = MagicMock(
            return_value={
                "127.0.0.1": {"aliases": ["localhost"]},
                cur_ip: {"aliases": [hostname]},
            }
        )
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_list)
            assert ret["result"] is True
            assert "Added host {0} ({1})".format(hostname, ip_list[0]) in ret["comment"]
            assert "Added host {0} ({1})".format(hostname, ip_list[1]) in ret["comment"]
            assert ret["changes"] == {
                "added": {ip_list[0]: [hostname], ip_list[1]: [hostname]},
            }, ret["changes"]
            expected = sorted([call(x, hostname) for x in ip_list])
            assert sorted(add_host.mock_calls) == expected, add_host.mock_calls
            assert rm_host.mock_calls == [], rm_host.mock_calls

        # Case 4a: Repeat the above with clean=True
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_list, clean=True)
            assert ret["result"] is True
            assert "Added host {0} ({1})".format(hostname, ip_list[0]) in ret["comment"]
            assert "Added host {0} ({1})".format(hostname, ip_list[1]) in ret["comment"]
            assert "Removed host {0} ({1})".format(hostname, cur_ip) in ret["comment"]
            assert ret["changes"] == {
                "added": {ip_list[0]: [hostname], ip_list[1]: [hostname]},
                "removed": {cur_ip: [hostname]},
            }, ret["changes"]
            expected = sorted([call(x, hostname) for x in ip_list])
            assert sorted(add_host.mock_calls) == expected, add_host.mock_calls
            expected = [call(cur_ip, hostname)]
            assert rm_host.mock_calls == expected, rm_host.mock_calls

        # Case 5: Multiple IP addresses passed to the state. One of them
        # matches, the other does not. There is also a non-matching IP that
        # must be removed.
        cur_ip = "1.2.3.4"
        list_hosts = MagicMock(
            return_value={
                "127.0.0.1": {"aliases": ["localhost"]},
                cur_ip: {"aliases": [hostname]},
                ip_list[0]: {"aliases": [hostname]},
            }
        )
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_list)
            assert ret["result"] is True
            assert "Added host {0} ({1})".format(hostname, ip_list[1]) in ret["comment"]
            assert ret["changes"] == {"added": {ip_list[1]: [hostname]}}, ret["changes"]
            expected = [call(ip_list[1], hostname)]
            assert add_host.mock_calls == expected, add_host.mock_calls
            assert rm_host.mock_calls == [], rm_host.mock_calls

        # Case 5a: Repeat the above with clean=True
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_list, clean=True)
            assert ret["result"] is True
            assert "Added host {0} ({1})".format(hostname, ip_list[1]) in ret["comment"]
            assert "Removed host {0} ({1})".format(hostname, cur_ip) in ret["comment"]
            assert ret["changes"] == {
                "added": {ip_list[1]: [hostname]},
                "removed": {cur_ip: [hostname]},
            }, ret["changes"]
            expected = [call(ip_list[1], hostname)]
            assert add_host.mock_calls == expected, add_host.mock_calls
            expected = [call(cur_ip, hostname)]
            assert rm_host.mock_calls == expected, rm_host.mock_calls

        # Case 6: Single IP address passed to the state, which matches the
        # current configuration for that hostname. No changes should be made.
        list_hosts = MagicMock(return_value={ip_str: {"aliases": [hostname]}})
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_str)
            assert ret["result"] is True
            assert (
                ret["comment"]
                == "Host {0} ({1}) already present".format(hostname, ip_str)
                in ret["comment"]
            )
            assert ret["changes"] == {}, ret["changes"]
            assert add_host.mock_calls == [], add_host.mock_calls
            assert rm_host.mock_calls == [], rm_host.mock_calls

        # Case 7: Multiple IP addresses passed to the state, which both match
        # the current configuration for that hostname. No changes should be
        # made.
        list_hosts = MagicMock(
            return_value={
                ip_list[0]: {"aliases": [hostname]},
                ip_list[1]: {"aliases": [hostname]},
            }
        )
        add_host.reset_mock()
        rm_host.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": add_host,
                "hosts.rm_host": rm_host,
            },
        ):
            ret = host.present(hostname, ip_list)
            assert ret["result"] is True
            assert (
                "Host {0} ({1}) already present".format(hostname, ip_list[0])
                in ret["comment"]
            )
            assert (
                "Host {0} ({1}) already present".format(hostname, ip_list[1])
                in ret["comment"]
            )
            assert ret["changes"] == {}, ret["changes"]
            assert add_host.mock_calls == [], add_host.mock_calls
            assert rm_host.mock_calls == [], rm_host.mock_calls

        # Case 8: Passing a comment to host.present with multiple IPs
        list_hosts = MagicMock(return_value={"127.0.0.1": {"aliases": ["localhost"]}})
        self.add_host_mock.reset_mock()
        self.rm_host_mock.reset_mock()
        self.set_comment_mock.reset_mock()
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": self.add_host_mock,
                "hosts.rm_host": self.rm_host_mock,
                "hosts.set_comment": self.set_comment_mock,
            },
        ):
            ret = host.present(self.hostname, self.ip_list, comment="A comment")
            assert ret["result"] is True
            assert (
                "Added host {0} ({1})".format(self.hostname, self.ip_list[0])
                in ret["comment"]
            )
            assert (
                "Added host {0} ({1})".format(self.hostname, self.ip_list[1])
                in ret["comment"]
            )
            assert (
                "Set comment for host {0} (A comment)".format(self.ip_list[0])
                in ret["comment"]
            )
            assert (
                "Set comment for host {0} (A comment)".format(self.ip_list[1])
                in ret["comment"]
            )
            assert ret["changes"] == {
                "added": {
                    self.ip_list[0]: [self.hostname],
                    self.ip_list[1]: [self.hostname],
                },
                "comment_added": {
                    self.ip_list[0]: ["A comment"],
                    self.ip_list[1]: ["A comment"],
                },
            }, ret["changes"]
            expected = sorted([call(x, self.hostname) for x in self.ip_list])
            assert (
                sorted(self.add_host_mock.mock_calls) == expected
            ), self.add_host_mock.mock_calls

            expected = sorted([call(x, "A comment") for x in self.ip_list])
            assert (
                sorted(self.set_comment_mock.mock_calls) == expected
            ), self.set_comment_mock.mock_calls

            assert self.rm_host_mock.mock_calls == [], self.rm_host_mock.mock_calls

    def test_host_present_should_return_True_if_test_and_no_changes(self):
        expected = {
            "comment": "Host {} ({}) already present".format(
                self.hostname, self.ip_list[0],
            ),
            "changes": {},
            "name": self.hostname,
            "result": True,
        }
        list_hosts = MagicMock(
            return_value={self.ip_list[0]: {"aliases": [self.hostname]}},
        )
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": self.add_host_mock,
                "hosts.rm_host": self.rm_host_mock,
            },
        ):
            with patch.dict(host.__opts__, {"test": True}):
                ret = host.present(self.hostname, self.ip_list[:1])

                self.assertDictEqual(ret, expected)

    def test_host_present_should_return_None_if_test_and_adding(self):
        expected = {
            "comment": "\n".join(
                ["Host {} ({}) already present", "Host {} ({}) would be added"]
            ).format(self.hostname, self.ip_list[0], self.hostname, self.ip_list[1],),
            "changes": {"added": {self.ip_list[1]: [self.hostname]}},
            "name": self.hostname,
            "result": None,
        }
        list_hosts = MagicMock(
            return_value={self.ip_list[0]: {"aliases": [self.hostname]}},
        )
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": list_hosts,
                "hosts.add_host": self.add_host_mock,
                "hosts.rm_host": self.rm_host_mock,
            },
        ):
            with patch.dict(host.__opts__, {"test": True}):
                ret = host.present(self.hostname, self.ip_list)
                self.assertDictEqual(ret, expected)

    def test_host_present_should_return_None_if_test_and_removing(self):
        expected = {
            "comment": "\n".join(
                ["Host {} ({}) already present", "Host {} ({}) would be removed"]
            ).format(self.hostname, self.ip_list[0], self.hostname, self.ip_list[1],),
            "changes": {"removed": {self.ip_list[1]: [self.hostname]}},
            "name": self.hostname,
            "result": None,
        }
        with patch.dict(
            host.__salt__,
            {
                "hosts.list_hosts": self.list_hosts_mock,
                "hosts.add_host": self.add_host_mock,
                "hosts.rm_host": self.rm_host_mock,
            },
        ):
            with patch.dict(host.__opts__, {"test": True}):
                ret = host.present(self.hostname, self.ip_list[:1], clean=True)
                self.assertDictEqual(ret, expected)

    def test_absent(self):
        """
        Test to ensure that the named host is absent
        """
        ret = {
            "changes": {},
            "comment": "Host salt (127.0.0.1) already absent",
            "name": "salt",
            "result": True,
        }

        mock = MagicMock(return_value=False)
        with patch.dict(host.__salt__, {"hosts.has_pair": mock}):
            self.assertDictEqual(host.absent("salt", "127.0.0.1"), ret)

    def test_only_already(self):
        """
        Test only() when the state hasn't changed
        """
        expected = {
            "name": "127.0.1.1",
            "changes": {},
            "result": True,
            "comment": 'IP address 127.0.1.1 already set to "foo.bar"',
        }
        mock1 = MagicMock(return_value=["foo.bar"])
        with patch.dict(host.__salt__, {"hosts.get_alias": mock1}):
            mock2 = MagicMock(return_value=True)
            with patch.dict(host.__salt__, {"hosts.set_host": mock2}):
                with patch.dict(host.__opts__, {"test": False}):
                    self.assertDictEqual(expected, host.only("127.0.1.1", "foo.bar"))

    def test_only_dryrun(self):
        """
        Test only() when state would change, but it's a dry run
        """
        expected = {
            "name": "127.0.1.1",
            "changes": {},
            "result": None,
            "comment": 'Would change 127.0.1.1 from "foo.bar" to "foo.bar foo"',
        }
        mock1 = MagicMock(return_value=["foo.bar"])
        with patch.dict(host.__salt__, {"hosts.get_alias": mock1}):
            mock2 = MagicMock(return_value=True)
            with patch.dict(host.__salt__, {"hosts.set_host": mock2}):
                with patch.dict(host.__opts__, {"test": True}):
                    self.assertDictEqual(
                        expected, host.only("127.0.1.1", ["foo.bar", "foo"])
                    )

    def test_only_fail(self):
        """
        Test only() when state change fails
        """
        expected = {
            "name": "127.0.1.1",
            "changes": {},
            "result": False,
            "comment": "hosts.set_host failed to change 127.0.1.1"
            + ' from "foo.bar" to "foo.bar foo"',
        }
        mock = MagicMock(return_value=["foo.bar"])
        with patch.dict(host.__salt__, {"hosts.get_alias": mock}):
            mock = MagicMock(return_value=False)
            with patch.dict(host.__salt__, {"hosts.set_host": mock}):
                with patch.dict(host.__opts__, {"test": False}):
                    self.assertDictEqual(
                        expected, host.only("127.0.1.1", ["foo.bar", "foo"])
                    )

    def test_only_success(self):
        """
        Test only() when state successfully changes
        """
        expected = {
            "name": "127.0.1.1",
            "changes": {"127.0.1.1": {"old": "foo.bar", "new": "foo.bar foo"}},
            "result": True,
            "comment": "successfully changed 127.0.1.1"
            + ' from "foo.bar" to "foo.bar foo"',
        }
        mock = MagicMock(return_value=["foo.bar"])
        with patch.dict(host.__salt__, {"hosts.get_alias": mock}):
            mock = MagicMock(return_value=True)
            with patch.dict(host.__salt__, {"hosts.set_host": mock}):
                with patch.dict(host.__opts__, {"test": False}):
                    self.assertDictEqual(
                        expected, host.only("127.0.1.1", ["foo.bar", "foo"])
                    )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.pagerduty as pagerduty

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PagerdutyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.pagerduty
    """

    def setup_loader_modules(self):
        return {pagerduty: {}}

    # 'create_event' function tests: 1

    def test_create_event(self):
        """
        Test to create an event on the PagerDuty service.
        """
        name = "This is a server warning message"
        details = "This is a much more detailed message"
        service_key = "9abcd123456789efabcde362783cdbaf"
        profile = "my-pagerduty-account"

        ret = {"name": name, "result": None, "comment": "", "changes": {}}

        with patch.dict(pagerduty.__opts__, {"test": True}):
            comt = "Need to create event: {0}".format(name)
            ret.update({"comment": comt})
            self.assertDictEqual(
                pagerduty.create_event(name, details, service_key, profile), ret
            )

        with patch.dict(pagerduty.__opts__, {"test": False}):
            mock_t = MagicMock(return_value=True)
            with patch.dict(pagerduty.__salt__, {"pagerduty.create_event": mock_t}):
                comt = "Created event: {0}".format(name)
                ret.update({"comment": comt, "result": True})
                self.assertDictEqual(
                    pagerduty.create_event(name, details, service_key, profile), ret
                )



# -*- coding: utf-8 -*-

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

import collections
import textwrap

# Import Salt libs
import salt.renderers.yaml as yaml
from salt.ext import six

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class YAMLRendererTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {yaml: {}}

    def assert_unicode(self, value):
        """
        Make sure the entire data structure is unicode
        """
        if six.PY3:
            return
        if isinstance(value, six.string_types):
            if not isinstance(value, six.text_type):
                self.raise_error(value)
        elif isinstance(value, collections.Mapping):
            for k, v in six.iteritems(value):
                self.assert_unicode(k)
                self.assert_unicode(v)
        elif isinstance(value, collections.Iterable):
            for item in value:
                self.assert_unicode(item)

    def assert_matches(self, ret, expected):
        self.assertEqual(ret, expected)
        self.assert_unicode(ret)

    def test_yaml_render_string(self):
        data = "string"
        result = yaml.render(data)

        self.assertEqual(result, data)

    def test_yaml_render_unicode(self):
        data = "!!python/unicode python unicode string"
        result = yaml.render(data)

        self.assertEqual(result, "python unicode string")

    def test_yaml_render_old_unicode(self):
        config = {"use_yamlloader_old": True}
        with patch.dict(yaml.__opts__, config):  # pylint: disable=no-member
            self.assert_matches(
                yaml.render(
                    textwrap.dedent(
                        """\
                    foo:
                      a: Д
                      b: {'a': u'\\u0414'}"""
                    )
                ),
                {"foo": {"a": "\u0414", "b": {"a": "\u0414"}}},
            )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.win_network as win_network

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class WinNetworkTestCase(TestCase, LoaderModuleMockMixin):
    """
        Validate the nftables state
    """

    def setup_loader_modules(self):
        return {win_network: {}}

    def test_managed_missing_parameters(self):
        """
        Test to ensure that the named interface is configured properly.
        """
        ret = {
            "name": "salt",
            "changes": {},
            "result": False,
            "comment": "dns_proto must be one of the following: static, dhcp\n"
            "ip_proto must be one of the following: static, dhcp",
        }
        self.assertDictEqual(win_network.managed("salt"), ret)

    def test_managed_static_enabled_false(self):
        ret = {
            "name": "salt",
            "changes": {},
            "result": True,
            "comment": "Interface 'salt' is up to date (already disabled)",
        }
        mock_false = MagicMock(return_value=False)
        with patch.dict(win_network.__salt__, {"ip.is_enabled": mock_false}):
            self.assertDictEqual(
                win_network.managed(
                    "salt", dns_proto="static", ip_proto="static", enabled=False
                ),
                ret,
            )

    def test_managed_test_true(self):
        ret = {
            "name": "salt",
            "changes": {},
            "result": False,
            "comment": "Failed to enable interface 'salt' to make changes",
        }
        mock_false = MagicMock(return_value=False)
        with patch.dict(
            win_network.__salt__, {"ip.is_enabled": mock_false, "ip.enable": mock_false}
        ), patch.dict(win_network.__opts__, {"test": False}):
            self.assertDictEqual(
                win_network.managed("salt", dns_proto="static", ip_proto="static"), ret
            )

    def test_managed_validate_errors(self):
        ret = {
            "name": "salt",
            "changes": {},
            "result": False,
            "comment": "The following SLS configuration errors were "
            "detected:\n"
            "- First Error\n"
            "- Second Error",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=["First Error", "Second Error"])
        with patch.dict(
            win_network.__salt__, {"ip.is_enabled": mock_true}
        ), patch.object(win_network, "_validate", mock_validate):
            self.assertDictEqual(
                win_network.managed("salt", dns_proto="static", ip_proto="static"), ret
            )

    def test_managed_get_current_config_failed(self):
        ret = {
            "name": "salt",
            "changes": {},
            "result": False,
            "comment": "Unable to get current configuration for interface " "'salt'",
        }
        mock_true = MagicMock(return_value=True)
        mock_false = MagicMock(return_value=False)
        mock_validate = MagicMock(return_value=[])
        with patch.dict(
            win_network.__salt__,
            {"ip.is_enabled": mock_true, "ip.get_interface": mock_false},
        ), patch.object(win_network, "_validate", mock_validate):

            self.assertDictEqual(
                win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp"), ret
            )

    def test_managed_test_true_no_changes(self):
        ret = {
            "name": "salt",
            "changes": {},
            "result": True,
            "comment": "Interface 'salt' is up to date",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=[])
        mock_get_int = MagicMock(
            return_value={
                "DHCP enabled": "yes",
                "DNS servers configured through DHCP": "192.168.0.10",
            }
        )
        with patch.dict(
            win_network.__salt__,
            {"ip.is_enabled": mock_true, "ip.get_interface": mock_get_int},
        ), patch.dict(win_network.__opts__, {"test": True}), patch.object(
            win_network, "_validate", mock_validate
        ):
            self.assertDictEqual(
                win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp"), ret
            )

    def test_managed_test_true_changes(self):
        ret = {
            "name": "salt",
            "changes": {},
            "result": None,
            "comment": "The following changes will be made to interface "
            "'salt':\n"
            "- DNS protocol will be changed to: dhcp",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=[])
        mock_get_int = MagicMock(
            return_value={
                "DHCP enabled": "no",
                "Statically Configured DNS Servers": "192.168.0.10",
            }
        )
        with patch.dict(
            win_network.__salt__,
            {"ip.is_enabled": mock_true, "ip.get_interface": mock_get_int},
        ), patch.dict(win_network.__opts__, {"test": True}), patch.object(
            win_network, "_validate", mock_validate
        ):

            self.assertDictEqual(
                win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp"), ret
            )

    def test_managed_failed(self):
        ret = {
            "name": "salt",
            "changes": {},
            "result": False,
            "comment": "Failed to set desired configuration settings for "
            "interface 'salt'",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=[])
        mock_get_int = MagicMock(
            return_value={
                "DHCP enabled": "no",
                "Statically Configured DNS Servers": "192.168.0.10",
            }
        )
        with patch.dict(
            win_network.__salt__,
            {
                "ip.is_enabled": mock_true,
                "ip.get_interface": mock_get_int,
                "ip.set_dhcp_dns": mock_true,
                "ip.set_dhcp_ip": mock_true,
            },
        ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
            win_network, "_validate", mock_validate
        ):
            self.assertDictEqual(
                win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp"), ret
            )

    def test_managed(self):
        ret = {
            "name": "salt",
            "changes": {
                "DHCP enabled": {"new": "yes", "old": "no"},
                "DNS servers configured through DHCP": {
                    "new": "192.168.0.10",
                    "old": "",
                },
                "Statically Configured DNS Servers": {"new": "", "old": "192.168.0.10"},
            },
            "result": True,
            "comment": "Successfully updated configuration for interface " "'salt'",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=[])
        mock_get_int = MagicMock(
            side_effect=[
                {
                    "DHCP enabled": "no",
                    "Statically Configured DNS Servers": "192.168.0.10",
                },
                {
                    "DHCP enabled": "yes",
                    "DNS servers configured through DHCP": "192.168.0.10",
                },
            ]
        )
        with patch.dict(
            win_network.__salt__,
            {
                "ip.is_enabled": mock_true,
                "ip.get_interface": mock_get_int,
                "ip.set_dhcp_dns": mock_true,
                "ip.set_dhcp_ip": mock_true,
            },
        ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
            win_network, "_validate", mock_validate
        ):
            self.assertDictEqual(
                win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp"), ret
            )

    def test_managed_static_dns_clear(self):
        expected = {
            "name": "salt",
            "changes": {
                "Statically Configured DNS Servers": {
                    "new": "None",
                    "old": "192.168.0.10",
                }
            },
            "result": True,
            "comment": "Successfully updated configuration for " "interface 'salt'",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=[])
        mock_get_int = MagicMock(
            side_effect=[
                {
                    "DHCP enabled": "no",
                    "Statically Configured DNS Servers": "192.168.0.10",
                },
                {"DHCP enabled": "no", "Statically Configured DNS Servers": "None"},
            ]
        )
        with patch.dict(
            win_network.__salt__,
            {
                "ip.is_enabled": mock_true,
                "ip.get_interface": mock_get_int,
                "ip.set_static_dns": mock_true,
            },
        ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
            win_network, "_validate", mock_validate
        ):
            ret = win_network.managed(
                "salt", dns_proto="static", dns_servers=[], ip_proto="dhcp"
            )
            self.assertDictEqual(ret, expected)

    def test_managed_static_dns(self):
        expected = {
            "name": "salt",
            "changes": {
                "Statically Configured DNS Servers": {
                    "new": "192.168.0.10",
                    "old": "None",
                }
            },
            "result": True,
            "comment": "Successfully updated configuration for " "interface 'salt'",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=[])
        mock_get_int = MagicMock(
            side_effect=[
                {"DHCP enabled": "no", "Statically Configured DNS Servers": "None"},
                {
                    "DHCP enabled": "no",
                    "Statically Configured DNS Servers": "192.168.0.10",
                },
            ]
        )
        with patch.dict(
            win_network.__salt__,
            {
                "ip.is_enabled": mock_true,
                "ip.get_interface": mock_get_int,
                "ip.set_static_dns": mock_true,
            },
        ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
            win_network, "_validate", mock_validate
        ):
            ret = win_network.managed(
                "salt",
                dns_proto="static",
                dns_servers=["192.168.0.10"],
                ip_proto="dhcp",
            )
            self.assertDictEqual(ret, expected)

    def test_managed_static_dns_no_action(self):
        expected = {
            "name": "salt",
            "changes": {},
            "result": True,
            "comment": "Interface 'salt' is up to date",
        }
        mock_true = MagicMock(return_value=True)
        mock_validate = MagicMock(return_value=[])
        mock_get_int = MagicMock(
            return_value={
                "DHCP enabled": "no",
                "Statically Configured DNS Servers": "192.168.0.10",
            }
        )
        with patch.dict(
            win_network.__salt__,
            {
                "ip.is_enabled": mock_true,
                "ip.get_interface": mock_get_int,
                "ip.set_static_dns": mock_true,
            },
        ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
            win_network, "_validate", mock_validate
        ):
            # Don't pass dns_servers
            ret = win_network.managed("salt", dns_proto="static", ip_proto="dhcp")
            self.assertDictEqual(ret, expected)
            # Pass dns_servers=None
            ret = win_network.managed(
                "salt", dns_proto="static", dns_servers=None, ip_proto="dhcp"
            )
            self.assertDictEqual(ret, expected)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""
from __future__ import absolute_import, print_function, unicode_literals

import copy
import os
import pathlib
import shutil
import tempfile

import salt.fileclient
import salt.fileserver.roots as roots
import salt.utils.files
import salt.utils.hashutils
import salt.utils.platform
from tests.support.mixins import (
    AdaptedConfigurationTestCaseMixin,
    LoaderModuleMockMixin,
)
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf

UNICODE_FILENAME = "питон.txt"
UNICODE_DIRNAME = UNICODE_ENVNAME = "соль"


class RootsTest(TestCase, AdaptedConfigurationTestCaseMixin, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        self.opts = self.get_temp_config("master")
        empty_dir = os.path.join(RUNTIME_VARS.TMP_STATE_TREE, "empty_dir")
        if not os.path.isdir(empty_dir):
            os.makedirs(empty_dir)
        return {roots: {"__opts__": self.opts}}

    @classmethod
    def setUpClass(cls):
        cls.tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        full_path_to_file = os.path.join(RUNTIME_VARS.BASE_FILES, "testfile")
        shutil.copyfile(full_path_to_file, os.path.join(cls.tmp_dir, "testfile"))

    @classmethod
    def tearDownClass(cls):
        salt.utils.files.rm_rf(cls.tmp_dir)

    def tearDown(self):
        del self.opts

    def test_file_list(self):
        ret = roots.file_list({"saltenv": "base"})
        self.assertIn("testfile", ret)
        self.assertIn(UNICODE_FILENAME, ret)

    def test_find_file(self):
        ret = roots.find_file("testfile")
        self.assertEqual("testfile", ret["rel"])

        full_path_to_file = os.path.join(RUNTIME_VARS.BASE_FILES, "testfile")
        self.assertEqual(full_path_to_file, ret["path"])

    def test_serve_file(self):
        with patch.dict(roots.__opts__, {"file_buffer_size": 262144}):
            load = {
                "saltenv": "base",
                "path": os.path.join(self.tmp_dir, "testfile"),
                "loc": 0,
            }
            fnd = {"path": os.path.join(self.tmp_dir, "testfile"), "rel": "testfile"}
            ret = roots.serve_file(load, fnd)

            with salt.utils.files.fopen(
                os.path.join(RUNTIME_VARS.BASE_FILES, "testfile"), "rb"
            ) as fp_:
                data = fp_.read()

            self.assertDictEqual(ret, {"data": data, "dest": "testfile"})

    def test_envs(self):
        opts = {"file_roots": copy.copy(self.opts["file_roots"])}
        opts["file_roots"][UNICODE_ENVNAME] = opts["file_roots"]["base"]
        with patch.dict(roots.__opts__, opts):
            ret = roots.envs()
        self.assertIn("base", ret)
        self.assertIn(UNICODE_ENVNAME, ret)

    def test_file_hash(self):
        load = {
            "saltenv": "base",
            "path": os.path.join(self.tmp_dir, "testfile"),
        }
        fnd = {"path": os.path.join(self.tmp_dir, "testfile"), "rel": "testfile"}
        ret = roots.file_hash(load, fnd)

        # Hashes are different in Windows. May be how git translates line
        # endings
        with salt.utils.files.fopen(
            os.path.join(RUNTIME_VARS.BASE_FILES, "testfile"), "rb"
        ) as fp_:
            hsum = salt.utils.hashutils.sha256_digest(fp_.read())

        self.assertDictEqual(ret, {"hsum": hsum, "hash_type": "sha256"})

    def test_file_list_emptydirs(self):
        ret = roots.file_list_emptydirs({"saltenv": "base"})
        self.assertIn("empty_dir", ret)

    def test_file_list_with_slash(self):
        opts = {"file_roots": copy.copy(self.opts["file_roots"])}
        opts["file_roots"]["foo/bar"] = opts["file_roots"]["base"]
        load = {
            "saltenv": "foo/bar",
        }
        with patch.dict(roots.__opts__, opts):
            ret = roots.file_list(load)
        self.assertIn("testfile", ret)
        self.assertIn(UNICODE_FILENAME, ret)

    def test_dir_list(self):
        ret = roots.dir_list({"saltenv": "base"})
        self.assertIn("empty_dir", ret)
        self.assertIn(UNICODE_DIRNAME, ret)

    def test_symlink_list(self):
        source_sym = pathlib.Path(RUNTIME_VARS.TMP_BASEENV_STATE_TREE) / "source_sym"
        source_sym.write_text("")
        dest_sym = pathlib.Path(RUNTIME_VARS.TMP_BASEENV_STATE_TREE) / "dest_sym"
        dest_sym.symlink_to(str(source_sym))
        self.addCleanup(dest_sym.unlink)
        self.addCleanup(source_sym.unlink)
        ret = roots.symlink_list({"saltenv": "base"})
        self.assertDictEqual(ret, {"dest_sym": str(source_sym)})

    def test_dynamic_file_roots(self):
        dyn_root_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        top_sls = os.path.join(dyn_root_dir, "top.sls")
        with salt.utils.files.fopen(top_sls, "w") as fp_:
            fp_.write("{{saltenv}}:\n  '*':\n    - dynamo\n")
        dynamo_sls = os.path.join(dyn_root_dir, "dynamo.sls")
        with salt.utils.files.fopen(dynamo_sls, "w") as fp_:
            fp_.write("foo:\n  test.nop\n")
        opts = {"file_roots": copy.copy(self.opts["file_roots"])}
        opts["file_roots"]["__env__"] = [dyn_root_dir]
        with patch.dict(roots.__opts__, opts):
            ret1 = roots.find_file("dynamo.sls", "dyn")
            ret2 = roots.file_list({"saltenv": "dyn"})
        self.assertEqual("dynamo.sls", ret1["rel"])
        self.assertIn("top.sls", ret2)
        self.assertIn("dynamo.sls", ret2)

    @skipIf(
        salt.utils.platform.is_windows(),
        "Windows does not support this master function",
    )
    def test_update_no_change(self):
        # process all changes that have happen
        # changes will always take place the first time during testing
        ret = roots.update()
        self.assertTrue(ret["changed"])

        # check if no changes took place
        ret = roots.update()
        self.assertFalse(ret["changed"])
        self.assertEqual(ret["files"]["changed"], [])
        self.assertEqual(ret["files"]["removed"], [])
        self.assertEqual(ret["files"]["added"], [])



# -*- coding: utf-8 -*-
"""
    tests.support.pytest.helpers
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    PyTest helpers functions
"""
import os
import shutil
import tempfile
import textwrap
from contextlib import contextmanager

import pytest
import salt.utils.files
from tests.support.runtests import RUNTIME_VARS


@pytest.helpers.register
@contextmanager
def temp_directory(name=None):
    """
    This helper creates a temporary directory. It should be used as a context manager
    which returns the temporary directory path, and, once out of context, deletes it.

    Can be directly imported and used, or, it can be used as a pytest helper function if
    ``pytest-helpers-namespace`` is installed.

    .. code-block:: python

        import os
        import pytest

        def test_blah():
            with pytest.helpers.temp_directory() as tpath:
                print(tpath)
                assert os.path.exists(tpath)

            assert not os.path.exists(tpath)
    """
    try:
        if name is not None:
            directory_path = os.path.join(RUNTIME_VARS.TMP, name)
        else:
            directory_path = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        if not os.path.isdir(directory_path):
            os.makedirs(directory_path)

        yield directory_path
    finally:
        shutil.rmtree(directory_path, ignore_errors=True)


@pytest.helpers.register
@contextmanager
def temp_file(name=None, contents=None, directory=None, strip_first_newline=True):
    """
    This helper creates a temporary file. It should be used as a context manager
    which returns the temporary file path, and, once out of context, deletes it.

    Can be directly imported and used, or, it can be used as a pytest helper function if
    ``pytest-helpers-namespace`` is installed.

    .. code-block:: python

        import os
        import pytest

        def test_blah():
            with pytest.helpers.temp_file("blah.txt") as tpath:
                print(tpath)
                assert os.path.exists(tpath)

            assert not os.path.exists(tpath)

    Args:
        name(str):
            The temporary file name
        contents(str):
            The contents of the temporary file
        directory(str):
            The directory where to create the temporary file. If ``None``, then ``RUNTIME_VARS.TMP``
            will be used.
        strip_first_newline(bool):
            Wether to strip the initial first new line char or not.
    """
    try:
        if directory is None:
            directory = RUNTIME_VARS.TMP

        if name is not None:
            file_path = os.path.join(directory, name)
        else:
            handle, file_path = tempfile.mkstemp(dir=directory)
            os.close(handle)

        file_directory = os.path.dirname(file_path)
        if not os.path.isdir(file_directory):
            os.makedirs(file_directory)

        if contents is not None:
            if contents:
                if contents.startswith("\n") and strip_first_newline:
                    contents = contents[1:]
                file_contents = textwrap.dedent(contents)
            else:
                file_contents = contents

            with salt.utils.files.fopen(file_path, "w") as wfh:
                wfh.write(file_contents)

        yield file_path

    finally:
        try:
            os.unlink(file_path)
        except OSError:
            # Already deleted
            pass


@pytest.helpers.register
def temp_state_file(name, contents, saltenv="base", strip_first_newline=True):
    """
    This helper creates a temporary state file. It should be used as a context manager
    which returns the temporary state file path, and, once out of context, deletes it.

    Can be directly imported and used, or, it can be used as a pytest helper function if
    ``pytest-helpers-namespace`` is installed.

    .. code-block:: python

        import os
        import pytest

        def test_blah():
            with pytest.helpers.temp_state_file("blah.sls") as tpath:
                print(tpath)
                assert os.path.exists(tpath)

            assert not os.path.exists(tpath)

    Depending on the saltenv, it will be created under ``RUNTIME_VARS.TMP_STATE_TREE`` or
    ``RUNTIME_VARS.TMP_PRODENV_STATE_TREE``.

    Args:
        name(str):
            The temporary state file name
        contents(str):
            The contents of the temporary file
        saltenv(str):
            The salt env to use. Either ``base`` or ``prod``
        strip_first_newline(bool):
            Wether to strip the initial first new line char or not.
    """

    if saltenv == "base":
        directory = RUNTIME_VARS.TMP_STATE_TREE
    elif saltenv == "prod":
        directory = RUNTIME_VARS.TMP_PRODENV_STATE_TREE
    else:
        raise RuntimeError(
            '"saltenv" can only be "base" or "prod", not "{}"'.format(saltenv)
        )
    return temp_file(
        name, contents, directory=directory, strip_first_newline=strip_first_newline
    )


@pytest.helpers.register
def temp_pillar_file(name, contents, saltenv="base", strip_first_newline=True):
    """
    This helper creates a temporary pillar file. It should be used as a context manager
    which returns the temporary pillar file path, and, once out of context, deletes it.

    Can be directly imported and used, or, it can be used as a pytest helper function if
    ``pytest-helpers-namespace`` is installed.

    .. code-block:: python

        import os
        import pytest

        def test_blah():
            with pytest.helpers.temp_pillar_file("blah.sls") as tpath:
                print(tpath)
                assert os.path.exists(tpath)

            assert not os.path.exists(tpath)

    Depending on the saltenv, it will be created under ``RUNTIME_VARS.TMP_PILLAR_TREE`` or
    ``RUNTIME_VARS.TMP_PRODENV_PILLAR_TREE``.

    Args:
        name(str):
            The temporary state file name
        contents(str):
            The contents of the temporary file
        saltenv(str):
            The salt env to use. Either ``base`` or ``prod``
        strip_first_newline(bool):
            Wether to strip the initial first new line char or not.
    """

    if saltenv == "base":
        directory = RUNTIME_VARS.TMP_PILLAR_TREE
    elif saltenv == "prod":
        directory = RUNTIME_VARS.TMP_PRODENV_PILLAR_TREE
    else:
        raise RuntimeError(
            '"saltenv" can only be "base" or "prod", not "{}"'.format(saltenv)
        )
    return temp_file(
        name, contents, directory=directory, strip_first_newline=strip_first_newline
    )


# Only allow star importing the functions defined in this module
__all__ = [
    name
    for (name, func) in locals().items()
    if getattr(func, "__module__", None) == __name__
]



# -*- coding: utf-8 -*-
"""
    :synopsis: Unit Tests for 'module.aptkernelpkg'
    :platform: Linux
    :maturity: develop
    versionadded:: 2018.3.0
"""
# pylint: disable=invalid-name,no-member

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
from salt.ext import six

try:
    # Import Salt Testing Libs
    from tests.support.mixins import LoaderModuleMockMixin
    from tests.support.unit import skipIf, TestCase
    from tests.support.mock import MagicMock, patch

    # Import Salt Libs
    import salt.states.kernelpkg as kernelpkg

    HAS_MODULES = True
except ImportError:
    HAS_MODULES = False

KERNEL_LIST = ["4.4.0-70-generic", "4.4.0-71-generic", "4.5.1-14-generic"]
STATE_NAME = "kernelpkg-test"


@skipIf(not HAS_MODULES, "Salt modules could not be loaded")
class KernelPkgTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.aptpkg
    """

    def setup_loader_modules(self):
        return {
            kernelpkg: {
                "__salt__": {
                    "system.reboot": MagicMock(return_value=None),
                    "kernelpkg.upgrade": MagicMock(
                        return_value={
                            "upgrades": {"kernel": {"old": "1.0.0", "new": "2.0.0"}}
                        }
                    ),
                    "kernelpkg.active": MagicMock(return_value=0),
                    "kernelpkg.latest_installed": MagicMock(return_value=0),
                }
            }
        }

    def test_latest_installed_with_changes(self):
        """
        Test - latest_installed when an upgrade is available
        """
        installed = MagicMock(return_value=KERNEL_LIST[:-1])
        upgrade = MagicMock(return_value=KERNEL_LIST[-1])
        with patch.dict(kernelpkg.__salt__, {"kernelpkg.list_installed": installed}):
            with patch.dict(
                kernelpkg.__salt__, {"kernelpkg.latest_available": upgrade}
            ):
                with patch.dict(kernelpkg.__opts__, {"test": False}):
                    kernelpkg.__salt__["kernelpkg.upgrade"].reset_mock()
                    ret = kernelpkg.latest_installed(name=STATE_NAME)
                    self.assertEqual(ret["name"], STATE_NAME)
                    self.assertTrue(ret["result"])
                    self.assertIsInstance(ret["changes"], dict)
                    self.assertIsInstance(ret["comment"], six.text_type)
                    self.assert_called_once(kernelpkg.__salt__["kernelpkg.upgrade"])

                with patch.dict(kernelpkg.__opts__, {"test": True}):
                    kernelpkg.__salt__["kernelpkg.upgrade"].reset_mock()
                    ret = kernelpkg.latest_installed(name=STATE_NAME)
                    self.assertEqual(ret["name"], STATE_NAME)
                    self.assertIsNone(ret["result"])
                    self.assertDictEqual(ret["changes"], {})
                    self.assertIsInstance(ret["comment"], six.text_type)
                    kernelpkg.__salt__["kernelpkg.upgrade"].assert_not_called()

    def test_latest_installed_at_latest(self):
        """
        Test - latest_installed when no upgrade is available
        """
        installed = MagicMock(return_value=KERNEL_LIST)
        upgrade = MagicMock(return_value=KERNEL_LIST[-1])
        with patch.dict(kernelpkg.__salt__, {"kernelpkg.list_installed": installed}):
            with patch.dict(
                kernelpkg.__salt__, {"kernelpkg.latest_available": upgrade}
            ):
                with patch.dict(kernelpkg.__opts__, {"test": False}):
                    ret = kernelpkg.latest_installed(name=STATE_NAME)
                    self.assertEqual(ret["name"], STATE_NAME)
                    self.assertTrue(ret["result"])
                    self.assertDictEqual(ret["changes"], {})
                    self.assertIsInstance(ret["comment"], six.text_type)
                    kernelpkg.__salt__["kernelpkg.upgrade"].assert_not_called()

                with patch.dict(kernelpkg.__opts__, {"test": True}):
                    ret = kernelpkg.latest_installed(name=STATE_NAME)
                    self.assertEqual(ret["name"], STATE_NAME)
                    self.assertTrue(ret["result"])
                    self.assertDictEqual(ret["changes"], {})
                    self.assertIsInstance(ret["comment"], six.text_type)
                    kernelpkg.__salt__["kernelpkg.upgrade"].assert_not_called()

    def test_latest_active_with_changes(self):
        """
        Test - latest_active when a new kernel is available
        """
        reboot = MagicMock(return_value=True)
        latest = MagicMock(return_value=1)
        with patch.dict(
            kernelpkg.__salt__,
            {"kernelpkg.needs_reboot": reboot, "kernelpkg.latest_installed": latest},
        ), patch.dict(kernelpkg.__opts__, {"test": False}):
            kernelpkg.__salt__["system.reboot"].reset_mock()
            ret = kernelpkg.latest_active(name=STATE_NAME)
            self.assertEqual(ret["name"], STATE_NAME)
            self.assertTrue(ret["result"])
            self.assertIsInstance(ret["changes"], dict)
            self.assertIsInstance(ret["comment"], six.text_type)
            self.assert_called_once(kernelpkg.__salt__["system.reboot"])

            with patch.dict(kernelpkg.__opts__, {"test": True}):
                kernelpkg.__salt__["system.reboot"].reset_mock()
                ret = kernelpkg.latest_active(name=STATE_NAME)
                self.assertEqual(ret["name"], STATE_NAME)
                self.assertIsNone(ret["result"])
                self.assertDictEqual(ret["changes"], {"kernel": {"new": 1, "old": 0}})
                self.assertIsInstance(ret["comment"], six.text_type)
                kernelpkg.__salt__["system.reboot"].assert_not_called()

    def test_latest_active_at_latest(self):
        """
        Test - latest_active when the newest kernel is already active
        """
        reboot = MagicMock(return_value=False)
        with patch.dict(kernelpkg.__salt__, {"kernelpkg.needs_reboot": reboot}):
            with patch.dict(kernelpkg.__opts__, {"test": False}):
                kernelpkg.__salt__["system.reboot"].reset_mock()
                ret = kernelpkg.latest_active(name=STATE_NAME)
                self.assertEqual(ret["name"], STATE_NAME)
                self.assertTrue(ret["result"])
                self.assertDictEqual(ret["changes"], {})
                self.assertIsInstance(ret["comment"], six.text_type)
                kernelpkg.__salt__["system.reboot"].assert_not_called()

            with patch.dict(kernelpkg.__opts__, {"test": True}):
                kernelpkg.__salt__["system.reboot"].reset_mock()
                ret = kernelpkg.latest_active(name=STATE_NAME)
                self.assertEqual(ret["name"], STATE_NAME)
                self.assertTrue(ret["result"])
                self.assertDictEqual(ret["changes"], {})
                self.assertIsInstance(ret["comment"], six.text_type)
                kernelpkg.__salt__["system.reboot"].assert_not_called()

    def test_latest_wait(self):
        """
        Test - latest_wait static results
        """
        ret = kernelpkg.latest_wait(name=STATE_NAME)
        self.assertEqual(ret["name"], STATE_NAME)
        self.assertTrue(ret["result"])
        self.assertDictEqual(ret["changes"], {})
        self.assertIsInstance(ret["comment"], six.text_type)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

# Import Python libs
from __future__ import absolute_import

# Import Salt Libs
import salt.modules.postfix as postfix

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PostfixTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.postfix
    """

    def setup_loader_modules(self):
        return {postfix: {}}

    def test_show_master(self):
        """
        Test for return a dict of active config values
        """
        with patch.object(postfix, "_parse_master", return_value=({"A": "a"}, ["b"])):
            self.assertDictEqual(postfix.show_master("path"), {"A": "a"})

    def test_set_master(self):
        """
        Test for set a single config value in the master.cf file
        """
        with patch.object(postfix, "_parse_master", return_value=({"A": "a"}, ["b"])):
            with patch.object(postfix, "_write_conf", return_value=None):
                self.assertTrue(postfix.set_master("a", "b"))

    def test_show_main(self):
        """
        Test for return a dict of active config values
        """
        with patch.object(postfix, "_parse_main", return_value=({"A": "a"}, ["b"])):
            self.assertDictEqual(postfix.show_main("path"), {"A": "a"})

    def test_set_main(self):
        """
        Test for set a single config value in the master.cf file
        """
        with patch.object(postfix, "_parse_main", return_value=({"A": "a"}, ["b"])):
            with patch.object(postfix, "_write_conf", return_value=None):
                self.assertTrue(postfix.set_main("key", "value"))

    def test_show_queue(self):
        """
        Test for show contents of the mail queue
        """
        with patch.dict(postfix.__salt__, {"cmd.run": MagicMock(return_value="A\nB")}):
            self.assertEqual(postfix.show_queue(), [])

        # Test if get an extra newline in the output
        with patch.dict(
            postfix.__salt__, {"cmd.run": MagicMock(return_value="A\nB\n")}
        ):
            self.assertEqual(postfix.show_queue(), [])

    def test_delete(self):
        """
        Test for delete message(s) from the mail queue
        """
        with patch.object(postfix, "show_queue", return_value={}):
            self.assertDictEqual(
                postfix.delete("queue_id"),
                {"result": False, "message": "No message in queue with ID queue_id"},
            )

        with patch.dict(
            postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            self.assertDictEqual(
                postfix.delete("ALL"),
                {"result": True, "message": "Successfully removed all messages"},
            )

    def test_hold(self):
        """
        Test for set held message(s) in the mail queue to unheld
        """
        with patch.object(postfix, "show_queue", return_value={}):
            self.assertDictEqual(
                postfix.hold("queue_id"),
                {"result": False, "message": "No message in queue with ID queue_id"},
            )

        with patch.dict(
            postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            self.assertDictEqual(
                postfix.hold("ALL"),
                {"result": True, "message": "Successfully placed all messages on hold"},
            )

    def test_unhold(self):
        """
        Test for put message(s) on hold from the mail queue
        """
        with patch.object(postfix, "show_queue", return_value={}):
            self.assertDictEqual(
                postfix.unhold("queue_id"),
                {"result": False, "message": "No message in queue with ID queue_id"},
            )

        with patch.dict(
            postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            self.assertDictEqual(
                postfix.unhold("ALL"),
                {"result": True, "message": "Successfully set all message as unheld"},
            )

    def test_requeue(self):
        """
        Test for requeue message(s) in the mail queue
        """
        with patch.object(postfix, "show_queue", return_value={}):
            self.assertDictEqual(
                postfix.requeue("queue_id"),
                {"result": False, "message": "No message in queue with ID queue_id"},
            )

        with patch.dict(
            postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            self.assertDictEqual(
                postfix.requeue("ALL"),
                {"result": True, "message": "Successfully requeued all messages"},
            )



# -*- coding: utf-8 -*-
"""
Helpers for testing man pages
"""
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import sys

# Import Salt libs
import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

# Import Salt Tesing libs
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


def install(rootdir):
    if not os.path.exists(rootdir):
        os.makedirs(rootdir)
    return __salt__["cmd.run_all"](
        [
            sys.executable,
            os.path.join(RUNTIME_VARS.CODE_DIR, "setup.py"),
            "install",
            "--root={0}".format(rootdir),
        ],
        redirect_stderr=True,
    )


def search(manpages, rootdir):
    manpage_fns = set(manpages)
    manpage_paths = {}
    for root, _, files in os.walk(rootdir):
        if not manpage_fns:
            # All manpages found, no need to keep walking
            break
        # Using list because we will be modifying the set during iteration
        for manpage_fn in list(manpage_fns):
            if manpage_fn in files:
                manpage_path = salt.utils.path.join(root, manpage_fn)
                manpage_paths[manpage_fn] = manpage_path
                manpage_fns.remove(manpage_fn)

    if manpage_fns:
        raise CommandExecutionError(
            "The following manpages were not found under {0}: {1}".format(
                rootdir, ", ".join(sorted(manpage_fns))
            )
        )

    failed = {}
    for manpage in sorted(manpages):
        with salt.utils.files.fopen(manpage_paths[manpage]) as fp_:
            contents = salt.utils.stringutils.to_unicode(fp_.read())
        # Check for search string in contents
        for search_string in manpages[manpage]:
            if search_string not in contents:
                failed.setdefault(manpage, []).append(
                    "No match for search string '{0}' found in {1}".format(
                        search_string, manpage_paths[manpage]
                    )
                )
        # Check for correct install dir
        path = "/man{0}/".format(manpage.rsplit(".", 1)[-1])
        if path not in manpage_paths[manpage]:
            failed.setdefault(manpage, []).append(
                "{0} not found in manpage path {1}".format(path, manpage_paths[manpage])
            )

    if failed:
        raise CommandExecutionError("One or more manpages failed", info=failed)

    return True



# -*- coding: utf-8 -*-
"""
QingCloud Cloud Module
======================

.. versionadded:: 2015.8.0

The QingCloud cloud module is used to control access to the QingCloud.
http://www.qingcloud.com/

Use of this module requires the ``access_key_id``, ``secret_access_key``,
``zone`` and ``key_filename`` parameter to be set.

Set up the cloud configuration at ``/etc/salt/cloud.providers`` or
``/etc/salt/cloud.providers.d/qingcloud.conf``:

.. code-block:: yaml

    my-qingcloud:
      driver: qingcloud
      access_key_id: AKIDMRTGYONNLTFFRBQJ
      secret_access_key: clYwH21U5UOmcov4aNV2V2XocaHCG3JZGcxEczFu
      zone: pek2
      key_filename: /path/to/your.pem

:depends: requests
"""

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import base64
import hmac
import logging
import pprint
import time
from hashlib import sha256

import salt.config as config
import salt.utils.cloud
import salt.utils.data
import salt.utils.json
from salt.exceptions import (
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudNotFound,
    SaltCloudSystemExit,
)

# Import Salt Libs
from salt.ext import six
from salt.ext.six.moves import range
from salt.ext.six.moves.urllib.parse import quote as _quote

# Import Third Party Libs
try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False


# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "qingcloud"

DEFAULT_QINGCLOUD_API_VERSION = 1
DEFAULT_QINGCLOUD_SIGNATURE_VERSION = 1


# Only load in this module if the qingcloud configurations are in place
def __virtual__():
    """
    Check for QingCloud configurations.
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__,
        __active_provider_name__ or __virtualname__,
        ("access_key_id", "secret_access_key", "zone", "key_filename"),
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(__virtualname__, {"requests": HAS_REQUESTS})


def _compute_signature(parameters, access_key_secret, method, path):
    """
    Generate an API request signature. Detailed document can be found at:

    https://docs.qingcloud.com/api/common/signature.html
    """
    parameters["signature_method"] = "HmacSHA256"

    string_to_sign = "{0}\n{1}\n".format(method.upper(), path)

    keys = sorted(parameters.keys())
    pairs = []
    for key in keys:
        val = six.text_type(parameters[key]).encode("utf-8")
        pairs.append(_quote(key, safe="") + "=" + _quote(val, safe="-_~"))
    qs = "&".join(pairs)
    string_to_sign += qs

    h = hmac.new(access_key_secret, digestmod=sha256)
    h.update(string_to_sign)

    signature = base64.b64encode(h.digest()).strip()

    return signature


def query(params=None):
    """
    Make a web call to QingCloud IaaS API.
    """
    path = "https://api.qingcloud.com/iaas/"

    access_key_id = config.get_cloud_config_value(
        "access_key_id", get_configured_provider(), __opts__, search_global=False
    )
    access_key_secret = config.get_cloud_config_value(
        "secret_access_key", get_configured_provider(), __opts__, search_global=False
    )

    # public interface parameters
    real_parameters = {
        "access_key_id": access_key_id,
        "signature_version": DEFAULT_QINGCLOUD_SIGNATURE_VERSION,
        "time_stamp": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        "version": DEFAULT_QINGCLOUD_API_VERSION,
    }

    # include action or function parameters
    if params:
        for key, value in params.items():
            if isinstance(value, list):
                for i in range(1, len(value) + 1):
                    if isinstance(value[i - 1], dict):
                        for sk, sv in value[i - 1].items():
                            if isinstance(sv, dict) or isinstance(sv, list):
                                sv = salt.utils.json.dumps(sv, separators=(",", ":"))
                            real_parameters["{0}.{1}.{2}".format(key, i, sk)] = sv
                    else:
                        real_parameters["{0}.{1}".format(key, i)] = value[i - 1]
            else:
                real_parameters[key] = value

    # Calculate the string for Signature
    signature = _compute_signature(real_parameters, access_key_secret, "GET", "/iaas/")
    real_parameters["signature"] = signature

    # print('parameters:')
    # pprint.pprint(real_parameters)

    request = requests.get(path, params=real_parameters, verify=False)

    # print('url:')
    # print(request.url)

    if request.status_code != 200:
        raise SaltCloudSystemExit(
            "An error occurred while querying QingCloud. HTTP Code: {0}  "
            "Error: '{1}'".format(request.status_code, request.text)
        )

    log.debug(request.url)

    content = request.text
    result = salt.utils.json.loads(content)

    # print('response:')
    # pprint.pprint(result)

    if result["ret_code"] != 0:
        raise SaltCloudSystemExit(pprint.pformat(result.get("message", {})))

    return result


def avail_locations(call=None):
    """
    Return a dict of all available locations on the provider with
    relevant data.

    CLI Examples:

    .. code-block:: bash

        salt-cloud --list-locations my-qingcloud
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    params = {
        "action": "DescribeZones",
    }
    items = query(params=params)

    result = {}
    for region in items["zone_set"]:
        result[region["zone_id"]] = {}
        for key in region:
            result[region["zone_id"]][key] = six.text_type(region[key])

    return result


def _get_location(vm_=None):
    """
    Return the VM's location. Used by create().
    """
    locations = avail_locations()

    vm_location = six.text_type(
        config.get_cloud_config_value("zone", vm_, __opts__, search_global=False)
    )

    if not vm_location:
        raise SaltCloudNotFound("No location specified for this VM.")

    if vm_location in locations:
        return vm_location

    raise SaltCloudNotFound(
        "The specified location, '{0}', could not be found.".format(vm_location)
    )


def _get_specified_zone(kwargs=None, provider=None):
    if provider is None:
        provider = get_configured_provider()

    if isinstance(kwargs, dict):
        zone = kwargs.get("zone", None)
        if zone is not None:
            return zone

    zone = provider["zone"]
    return zone


def avail_images(kwargs=None, call=None):
    """
    Return a list of the images that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud --list-images my-qingcloud
        salt-cloud -f avail_images my-qingcloud zone=gd1
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    params = {
        "action": "DescribeImages",
        "provider": "system",
        "zone": _get_specified_zone(kwargs, get_configured_provider()),
    }
    items = query(params=params)

    result = {}
    for image in items["image_set"]:
        result[image["image_id"]] = {}
        for key in image:
            result[image["image_id"]][key] = image[key]

    return result


def _get_image(vm_):
    """
    Return the VM's image. Used by create().
    """
    images = avail_images()
    vm_image = six.text_type(
        config.get_cloud_config_value("image", vm_, __opts__, search_global=False)
    )

    if not vm_image:
        raise SaltCloudNotFound("No image specified for this VM.")

    if vm_image in images:
        return vm_image

    raise SaltCloudNotFound(
        "The specified image, '{0}', could not be found.".format(vm_image)
    )


def show_image(kwargs, call=None):
    """
    Show the details from QingCloud concerning an image.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f show_image my-qingcloud image=trustysrvx64c
        salt-cloud -f show_image my-qingcloud image=trustysrvx64c,coreos4
        salt-cloud -f show_image my-qingcloud image=trustysrvx64c zone=ap1
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The show_images function must be called with " "-f or --function"
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    images = kwargs["image"]
    images = images.split(",")

    params = {
        "action": "DescribeImages",
        "images": images,
        "zone": _get_specified_zone(kwargs, get_configured_provider()),
    }

    items = query(params=params)

    if not items["image_set"]:
        raise SaltCloudNotFound("The specified image could not be found.")

    result = {}
    for image in items["image_set"]:
        result[image["image_id"]] = {}
        for key in image:
            result[image["image_id"]][key] = image[key]

    return result


# QingCloud doesn't provide an API of geting instance sizes
QINGCLOUD_SIZES = {
    "pek2": {
        "c1m1": {"cpu": 1, "memory": "1G"},
        "c1m2": {"cpu": 1, "memory": "2G"},
        "c1m4": {"cpu": 1, "memory": "4G"},
        "c2m2": {"cpu": 2, "memory": "2G"},
        "c2m4": {"cpu": 2, "memory": "4G"},
        "c2m8": {"cpu": 2, "memory": "8G"},
        "c4m4": {"cpu": 4, "memory": "4G"},
        "c4m8": {"cpu": 4, "memory": "8G"},
        "c4m16": {"cpu": 4, "memory": "16G"},
    },
    "pek1": {
        "small_b": {"cpu": 1, "memory": "1G"},
        "small_c": {"cpu": 1, "memory": "2G"},
        "medium_a": {"cpu": 2, "memory": "2G"},
        "medium_b": {"cpu": 2, "memory": "4G"},
        "medium_c": {"cpu": 2, "memory": "8G"},
        "large_a": {"cpu": 4, "memory": "4G"},
        "large_b": {"cpu": 4, "memory": "8G"},
        "large_c": {"cpu": 4, "memory": "16G"},
    },
}
QINGCLOUD_SIZES["ap1"] = QINGCLOUD_SIZES["pek2"]
QINGCLOUD_SIZES["gd1"] = QINGCLOUD_SIZES["pek2"]


def avail_sizes(kwargs=None, call=None):
    """
    Return a list of the instance sizes that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud --list-sizes my-qingcloud
        salt-cloud -f avail_sizes my-qingcloud zone=pek2
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    zone = _get_specified_zone(kwargs, get_configured_provider())

    result = {}
    for size_key in QINGCLOUD_SIZES[zone]:
        result[size_key] = {}
        for attribute_key in QINGCLOUD_SIZES[zone][size_key]:
            result[size_key][attribute_key] = QINGCLOUD_SIZES[zone][size_key][
                attribute_key
            ]

    return result


def _get_size(vm_):
    """
    Return the VM's size. Used by create().
    """
    sizes = avail_sizes()

    vm_size = six.text_type(
        config.get_cloud_config_value("size", vm_, __opts__, search_global=False)
    )

    if not vm_size:
        raise SaltCloudNotFound("No size specified for this instance.")

    if vm_size in sizes.keys():
        return vm_size

    raise SaltCloudNotFound(
        "The specified size, '{0}', could not be found.".format(vm_size)
    )


def _show_normalized_node(full_node):
    """
    Normalize the QingCloud instance data. Used by list_nodes()-related
    functions.
    """
    public_ips = full_node.get("eip", [])
    if public_ips:
        public_ip = public_ips["eip_addr"]
        public_ips = [
            public_ip,
        ]

    private_ips = []
    for vxnet in full_node.get("vxnets", []):
        private_ip = vxnet.get("private_ip", None)
        if private_ip:
            private_ips.append(private_ip)

    normalized_node = {
        "id": full_node["instance_id"],
        "image": full_node["image"]["image_id"],
        "size": full_node["instance_type"],
        "state": full_node["status"],
        "private_ips": private_ips,
        "public_ips": public_ips,
    }

    return normalized_node


def list_nodes_full(call=None):
    """
    Return a list of the instances that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -F my-qingcloud
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    zone = _get_specified_zone()

    params = {
        "action": "DescribeInstances",
        "zone": zone,
        "status": ["pending", "running", "stopped", "suspended"],
    }
    items = query(params=params)

    log.debug("Total %s instances found in zone %s", items["total_count"], zone)

    result = {}

    if items["total_count"] == 0:
        return result

    for node in items["instance_set"]:
        normalized_node = _show_normalized_node(node)
        node.update(normalized_node)

        result[node["instance_id"]] = node

    provider = __active_provider_name__ or "qingcloud"
    if ":" in provider:
        comps = provider.split(":")
        provider = comps[0]

    __opts__["update_cachedir"] = True
    __utils__["cloud.cache_node_list"](result, provider, __opts__)

    return result


def list_nodes(call=None):
    """
    Return a list of the instances that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -Q my-qingcloud
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    nodes = list_nodes_full()

    ret = {}
    for instance_id, full_node in nodes.items():
        ret[instance_id] = {
            "id": full_node["id"],
            "image": full_node["image"],
            "size": full_node["size"],
            "state": full_node["state"],
            "public_ips": full_node["public_ips"],
            "private_ips": full_node["private_ips"],
        }

    return ret


def list_nodes_min(call=None):
    """
    Return a list of the instances that are on the provider. Only a list of
    instances names, and their state, is returned.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f list_nodes_min my-qingcloud
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The list_nodes_min function must be called with -f or --function."
        )

    nodes = list_nodes_full()

    result = {}
    for instance_id, full_node in nodes.items():
        result[instance_id] = {
            "name": full_node["instance_name"],
            "status": full_node["status"],
        }

    return result


def list_nodes_select(call=None):
    """
    Return a list of the instances that are on the provider, with selected
    fields.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -S my-qingcloud
    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full("function"), __opts__["query.selection"], call,
    )


def show_instance(instance_id, call=None, kwargs=None):
    """
    Show the details from QingCloud concerning an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a show_instance i-2f733r5n
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    params = {
        "action": "DescribeInstances",
        "instances.1": instance_id,
        "zone": _get_specified_zone(kwargs=None, provider=get_configured_provider()),
    }
    items = query(params=params)

    if items["total_count"] == 0:
        raise SaltCloudNotFound(
            "The specified instance, '{0}', could not be found.".format(instance_id)
        )

    full_node = items["instance_set"][0]
    normalized_node = _show_normalized_node(full_node)
    full_node.update(normalized_node)

    result = full_node

    return result


def _query_node_data(instance_id):
    data = show_instance(instance_id, call="action")

    if not data:
        return False

    if data.get("private_ips", []):
        return data


def create(vm_):
    """
    Create a single instance from a data dict.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -p qingcloud-ubuntu-c1m1 hostname1
        salt-cloud -m /path/to/mymap.sls -P
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                __active_provider_name__ or "qingcloud",
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{0}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    log.info("Creating Cloud VM %s", vm_["name"])

    # params
    params = {
        "action": "RunInstances",
        "instance_name": vm_["name"],
        "zone": _get_location(vm_),
        "instance_type": _get_size(vm_),
        "image_id": _get_image(vm_),
        "vxnets.1": vm_["vxnets"],
        "login_mode": vm_["login_mode"],
        "login_keypair": vm_["login_keypair"],
    }

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{0}/requesting".format(vm_["name"]),
        args={
            "kwargs": __utils__["cloud.filter_event"](
                "requesting", params, list(params)
            ),
        },
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    result = query(params)
    new_instance_id = result["instances"][0]

    try:
        data = salt.utils.cloud.wait_for_ip(
            _query_node_data,
            update_args=(new_instance_id,),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_["name"])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(six.text_type(exc))

    private_ip = data["private_ips"][0]

    log.debug("VM %s is now running", private_ip)

    vm_["ssh_host"] = private_ip

    # The instance is booted and accessible, let's Salt it!
    __utils__["cloud.bootstrap"](vm_, __opts__)

    log.info("Created Cloud VM '%s'", vm_["name"])

    log.debug("'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{0}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return data


def script(vm_):
    """
    Return the script deployment object.
    """
    deploy_script = salt.utils.cloud.os_script(
        config.get_cloud_config_value("script", vm_, __opts__),
        vm_,
        __opts__,
        salt.utils.cloud.salt_config_to_yaml(
            salt.utils.cloud.minion_config(__opts__, vm_)
        ),
    )

    return deploy_script


def start(instance_id, call=None):
    """
    Start an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a start i-2f733r5n
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Starting instance %s", instance_id)

    params = {
        "action": "StartInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
    }
    result = query(params)

    return result


def stop(instance_id, force=False, call=None):
    """
    Stop an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a stop i-2f733r5n
        salt-cloud -a stop i-2f733r5n force=True
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Stopping instance %s", instance_id)

    params = {
        "action": "StopInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
        "force": int(force),
    }
    result = query(params)

    return result


def reboot(instance_id, call=None):
    """
    Reboot an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a reboot i-2f733r5n
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Rebooting instance %s", instance_id)

    params = {
        "action": "RestartInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
    }
    result = query(params)

    return result


def destroy(instance_id, call=None):
    """
    Destroy an instance.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a destroy i-2f733r5n
        salt-cloud -d i-2f733r5n
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, " "-a or --action."
        )

    instance_data = show_instance(instance_id, call="action")
    name = instance_data["instance_name"]

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        "salt/cloud/{0}/destroying".format(name),
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    params = {
        "action": "TerminateInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
    }
    result = query(params)

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        "salt/cloud/{0}/destroyed".format(name),
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return result



# -*- coding: utf-8 -*-

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import os

import salt.utils.boto3mod as boto3mod

# Import Salt libs
import salt.utils.botomod as botomod
from salt.exceptions import SaltInvocationError
from salt.ext import six
from salt.utils.versions import LooseVersion

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf

# Import 3rd-party libs
# pylint: disable=import-error
try:
    import boto

    boto.ENDPOINTS_PATH = os.path.join(
        RUNTIME_VARS.TESTS_DIR, "unit/files/endpoints.json"
    )
    import boto.exception
    from boto.exception import BotoServerError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

try:
    import boto3

    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False

try:
    from moto import mock_ec2

    HAS_MOTO = True
except ImportError:
    HAS_MOTO = False

    def mock_ec2(self):
        """
        if the mock_ec2 function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows unit tests to use the @mock_ec2 decorator
        without a "NameError: name 'mock_ec2' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function


required_boto_version = "2.0.0"
required_boto3_version = "1.2.1"
region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}

service = "ec2"
resource_name = "test-instance"
resource_id = "i-a1b2c3"


error_body = """
<Response>
    <Errors>
         <Error>
           <Code>Error code text</Code>
           <Message>Error message</Message>
         </Error>
    </Errors>
    <RequestID>request ID</RequestID>
</Response>
"""

no_error_body = """
<Response>
    <Errors />
    <RequestID>request ID</RequestID>
</Response>
"""


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif LooseVersion(boto.__version__) < LooseVersion(required_boto_version):
        return False
    else:
        return True


def _has_required_boto3():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    try:
        if not HAS_BOTO3:
            return False
        elif LooseVersion(boto3.__version__) < LooseVersion(required_boto3_version):
            return False
        else:
            return True
    except AttributeError as exc:
        if "has no attribute '__version__'" not in six.text_type(exc):
            raise
        return False


def _has_required_moto():
    """
    Returns True/False boolean depending on if Moto is installed and correct
    version.
    """
    if not HAS_MOTO:
        return False
    else:
        import pkg_resources

        if LooseVersion(pkg_resources.get_distribution("moto").version) < LooseVersion(
            "0.3.7"
        ):
            return False
        return True


class BotoUtilsTestCaseBase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        module_globals = {
            "__salt__": {"config.option": MagicMock(return_value="dummy_opt")}
        }
        return {botomod: module_globals, boto3mod: module_globals}


class BotoUtilsCacheIdTestCase(BotoUtilsTestCaseBase):
    def test_set_and_get_with_no_auth_params(self):
        botomod.cache_id(service, resource_name, resource_id=resource_id)
        self.assertEqual(botomod.cache_id(service, resource_name), resource_id)

    def test_set_and_get_with_explicit_auth_params(self):
        botomod.cache_id(
            service, resource_name, resource_id=resource_id, **conn_parameters
        )
        self.assertEqual(
            botomod.cache_id(service, resource_name, **conn_parameters), resource_id
        )

    def test_set_and_get_with_different_region_returns_none(self):
        botomod.cache_id(
            service, resource_name, resource_id=resource_id, region="us-east-1"
        )
        self.assertEqual(
            botomod.cache_id(service, resource_name, region="us-west-2"), None
        )

    def test_set_and_get_after_invalidation_returns_none(self):
        botomod.cache_id(service, resource_name, resource_id=resource_id)
        botomod.cache_id(
            service, resource_name, resource_id=resource_id, invalidate=True
        )
        self.assertEqual(botomod.cache_id(service, resource_name), None)

    def test_partial(self):
        cache_id = botomod.cache_id_func(service)
        cache_id(resource_name, resource_id=resource_id)
        self.assertEqual(cache_id(resource_name), resource_id)


@skipIf(HAS_BOTO is False, "The boto module must be installed.")
@skipIf(HAS_MOTO is False, "The moto module must be installed.")
@skipIf(
    _has_required_boto() is False,
    "The boto module must be greater than"
    " or equal to version {0}".format(required_boto_version),
)
class BotoUtilsGetConnTestCase(BotoUtilsTestCaseBase):
    @mock_ec2
    def test_conn_is_cached(self):
        conn = botomod.get_connection(service, **conn_parameters)
        self.assertTrue(conn in botomod.__context__.values())

    @mock_ec2
    def test_conn_is_cache_with_profile(self):
        conn = botomod.get_connection(service, profile=conn_parameters)
        self.assertTrue(conn in botomod.__context__.values())

    @mock_ec2
    def test_get_conn_with_no_auth_params_raises_invocation_error(self):
        with patch(
            "boto.{0}.connect_to_region".format(service),
            side_effect=boto.exception.NoAuthHandlerFound(),
        ):
            with self.assertRaises(SaltInvocationError):
                botomod.get_connection(service)

    @mock_ec2
    def test_get_conn_error_raises_command_execution_error(self):
        with patch(
            "boto.{0}.connect_to_region".format(service),
            side_effect=BotoServerError(400, "Mocked error", body=error_body),
        ):
            with self.assertRaises(BotoServerError):
                botomod.get_connection(service)

    @mock_ec2
    def test_partial(self):
        get_conn = botomod.get_connection_func(service)
        conn = get_conn(**conn_parameters)
        self.assertTrue(conn in botomod.__context__.values())


@skipIf(HAS_BOTO is False, "The boto module must be installed.")
@skipIf(
    _has_required_boto() is False,
    "The boto module must be greater than"
    " or equal to version {0}".format(required_boto_version),
)
class BotoUtilsGetErrorTestCase(BotoUtilsTestCaseBase):
    def test_error_message(self):
        e = BotoServerError("400", "Mocked error", body=error_body)
        r = botomod.get_error(e)
        expected = {
            "aws": {
                "code": "Error code text",
                "message": "Error message",
                "reason": "Mocked error",
                "status": "400",
            },
            "message": "Mocked error: Error message",
        }
        self.assertEqual(r, expected)

    def test_exception_message_with_no_body(self):
        e = BotoServerError("400", "Mocked error")
        r = botomod.get_error(e)
        expected = {
            "aws": {"reason": "Mocked error", "status": "400"},
            "message": "Mocked error",
        }
        self.assertEqual(r, expected)

    def test_exception_message_with_no_error_in_body(self):
        e = BotoServerError("400", "Mocked error", body=no_error_body)
        r = botomod.get_error(e)
        expected = {
            "aws": {"reason": "Mocked error", "status": "400"},
            "message": "Mocked error",
        }
        self.assertEqual(r, expected)


@skipIf(HAS_BOTO is False, "The boto module must be installed.")
@skipIf(
    _has_required_boto() is False,
    "The boto module must be greater than"
    " or equal to version {0}".format(required_boto_version),
)
@skipIf(HAS_BOTO3 is False, "The boto3 module must be installed.")
@skipIf(
    _has_required_boto3() is False,
    "The boto3 module must be greater than"
    " or equal to version {0}".format(required_boto3_version),
)
class BotoBoto3CacheContextCollisionTest(BotoUtilsTestCaseBase):
    def test_context_conflict_between_boto_and_boto3_utils(self):
        botomod.assign_funcs(__name__, "ec2")
        boto3mod.assign_funcs(__name__, "ec2", get_conn_funcname="_get_conn3")

        boto_ec2_conn = botomod.get_connection(
            "ec2", region=region, key=secret_key, keyid=access_key
        )
        boto3_ec2_conn = boto3mod.get_connection(
            "ec2", region=region, key=secret_key, keyid=access_key
        )

        # These should *not* be the same object!
        self.assertNotEqual(id(boto_ec2_conn), id(boto3_ec2_conn))
